﻿// Bitground Software common helpers library
// http://www.bitground.com/
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Bitground.Common.Helpers
{
    public enum HashAlgorithm
    {
        SHA512 = 1
    }

    public enum HMACHashAlgorithm
    {
        HMACSHA512 = 1
    }

    public enum SymetricEncryptionAlgorithm
    {
        AES256 = 1
    }

    public abstract partial class Crypto
    {
        #region Hash

        public static HashAlgorithm DefaultHashAlgorithm = HashAlgorithm.SHA512;

        public int HashAlgorithmId
        {
            get
            {
                return (int)DefaultHashAlgorithm;
            }
        }

        public static byte[] Hash(HashAlgorithm alg, byte[] message)
        {
            switch (alg)
            {
                case HashAlgorithm.SHA512:
                    return SHA512.Hash(message);
                default:
                    throw new Exception("Invalid HashAlgorithm");
            }
        }

        public static string Hash(string message)
        {
            return
                Hex.ByteArrayToHexString(
                    Hash(
                        DefaultHashAlgorithm,
                        UTF8.StringToByteArray(message)));
        }

        public static string HashHex(string hexmessage)
        {
            return
                Hex.ByteArrayToHexString(
                    Hash(
                        DefaultHashAlgorithm,
                        Hex.HexStringToByteArray(hexmessage)));
        }

        #endregion

        #region HMAC Hash

        public static HMACHashAlgorithm DefaultHMACHashAlgorithm = HMACHashAlgorithm.HMACSHA512;

        public int HMACHashAlgorithmId
        {
            get
            {
                return (int)DefaultHMACHashAlgorithm;
            }
        }

        public static byte[] HMACHash(HMACHashAlgorithm alg, byte[] key, byte[] message)
        {
            switch (alg)
            {
                case HMACHashAlgorithm.HMACSHA512:
                    return HMACSHA512.Hash(key, message);
                default:
                    throw new Exception("Invalid HMACHashAlgorithm");
            }
        }

        public static string HMACHash(string key, string message)
        {
            return
                Hex.ByteArrayToHexString(
                    HMACHash(
                        DefaultHMACHashAlgorithm,
                        UTF8.StringToByteArray(key),
                        UTF8.StringToByteArray(message)));
        }

        #endregion

        #region Symetric encryption
        
        public static SymetricEncryptionAlgorithm DefaultSymetricEncryptionAlgorithm = SymetricEncryptionAlgorithm.AES256;

        public int SymetricEncryptionAlgorithmId
        {
            get
            {
                return (int)DefaultSymetricEncryptionAlgorithm;
            }
        }

        #region Utils

        // the output is multiple lines (64 characters each) base64 encoded strings
        public static string ByteArrayToBase64StringOpenSSLCompiant(byte[] ar)
        {
            string b64 = Base64.ByteArrayToBase64String(ar);

            string broken_b64 = b64.SubstringFlexible(0, 64) + "\n";

            for (int i = 1; i < 1 + b64.Length / 64; i++)
            {
                broken_b64 += 
                    b64.SubstringFlexible
                    (
                        i * 64, 
                        64
                    ) 
                    + 
                    (
                        ((1 + b64.Length / 64) == (i + 1)) ? 
                            ""
                            :
                            "\n"
                    );
            }

            return broken_b64;
        }

        #endregion

        #region Encryption

        public static byte[] Encrypt(SymetricEncryptionAlgorithm alg, byte[] key, byte[] message)
        {
            switch (alg)
            {
                case SymetricEncryptionAlgorithm.AES256:
                    return AES256.Encrypt(key, message);
                default:
                    throw new Exception("Invalid SymetricEncryptionAlgorithm");
            }
        }

        public static string Encrypt(string hexKey, string message)
        {
            return
                ByteArrayToBase64StringOpenSSLCompiant(
                    Encrypt(
                        DefaultSymetricEncryptionAlgorithm,
                        Hex.HexStringToByteArray(hexKey),
                        UTF8.StringToByteArray(message)));
        }

        public static string DemoEncrypt(string message)
        {
            string hexKey = "89EF01AB23CD";
            byte[] salt = new byte[] { 4, 5, 6, 7, 0, 1, 2, 3 };
            return
                ByteArrayToBase64StringOpenSSLCompiant(
                    AES256.EncryptCore(
                        salt,
                        Hex.HexStringToByteArray(hexKey),
                        UTF8.StringToByteArray(message)));
        }

        #endregion

        #region Decryption

        public static byte[] Decrypt(SymetricEncryptionAlgorithm alg, byte[] key, byte[] message)
        {
            switch (alg)
            {
                case SymetricEncryptionAlgorithm.AES256:
                    return AES256.Decrypt(key, message);
                default:
                    throw new Exception("Invalid SymetricEncryptionAlgorithm");
            }
        }

        public static string Decrypt(string hexKey, string base64Message)
        {
            return
                UTF8.ByteArrayToString(
                    Decrypt(
                        DefaultSymetricEncryptionAlgorithm,
                        Hex.HexStringToByteArray(hexKey),
                        Base64.Base64StringToByteArray(base64Message)));
        }

        #endregion

        #endregion
    }
}
