﻿// Bitground Software common helpers library
// http://www.bitground.com/
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;
using log4net;

namespace Bitground.Common.Helpers
{
    public abstract class Log
    {
        private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        public enum Logger
        {
            ConsoleAll = 1 << 0,
            ConsoleWarnings = 1 << 1,
            ConsoleErrors = 1 << 2,

            FileAll = 1 << 3,
            FileDebug = 1 << 4,
            FileTrace = 1 << 5,
            FileInfo = 1 << 6,
            FileWarnings = 1 << 7,
            FileErrors = 1 << 8,
            
            EventErrors = 1 << 9,
            
            MailErrors = 1 << 10
        }

        public static Logger LoggerFromString(string logger)
        {
            return (Logger)Enum.Parse(typeof(Logger), logger);
        }

        public class LoggerInit
        {
            public string appName;

            public string appInstance;

            public Logger loggersMask;

            public Console console;
            public class Console
            {
            }

            public File file;
            public class File
            {
                public string logFolder;
            };

            public Event eventl;
            public class Event
            {
                public string application;
                public string log;
            };

            public Mail mail;
            public class Mail
            {
                public string from;
                public string to;
                public string smtpHost;
                public int port;
            };
        }

        public static void Initialize(LoggerInit init)
        {
            if (null == init.appName)
            {
                throw new ArgumentException("The application (init.app) should pe specified");
            }
            if (null == init.appInstance)
            {
                throw new ArgumentException("The instance (init.instance) should pe specified");
            }

            init.appName = init.appName.ToUpper();
            init.appInstance = init.appInstance.ToUpper();

            string pattern = @"%utcdate{yyyyMMdd-HHmmss-fff} %-5level %5thread: %20.20type.%-30method [%20.20file:%-4.4line] %ndc%message%newline";
            string patternConsole = @"%utcdate{yyyyMMdd-HHmmss-fff} %-5level %5thread: %ndc%message%newline";

            string head = String.Format(
                            "--------------- [ Logger initialized at {0} ] ------------------------------------{1}",
                            Time.ToCanonicalAccurateFileString(Time.UTCNow),
                            Environment.NewLine);

            string foot = String.Format(
                            "--------------- [ Logger terminated ] ------------------------------------------------------------{0}{1}{2}",
                            Environment.NewLine,
                            Environment.NewLine,
                            Environment.NewLine);

            if (!init.file.logFolder.EndsWith(@"\"))
            {
                init.file.logFolder += @"\";
            }

            if (0 != (init.loggersMask & Logger.ConsoleAll))
            {
                var cl = new log4net.Appender.ColoredConsoleAppender()
                {
                    Layout = new log4net.Layout.PatternLayout(patternConsole)
                    {
                        IgnoresException = true
                    },
                    Threshold = log4net.Core.Level.All
                };
                cl.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(cl);
            }

            if (0 != (init.loggersMask & Logger.ConsoleWarnings))
            {
                var cl = new log4net.Appender.ColoredConsoleAppender()
                {
                    Layout = new log4net.Layout.PatternLayout(patternConsole)
                    {
                        IgnoresException = true
                    },
                    Threshold = log4net.Core.Level.Warn
                };
                cl.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(cl);
            }

            if (0 != (init.loggersMask & Logger.ConsoleErrors))
            {
                var cl = new log4net.Appender.ColoredConsoleAppender()
                {
                    Layout = new log4net.Layout.PatternLayout(patternConsole)
                    {
                        IgnoresException = true
                    },
                    Threshold = log4net.Core.Level.Error
                };
                cl.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(cl);
            }

            if ((0 != (init.loggersMask & Logger.FileAll)) ||
                (0 != (init.loggersMask & Logger.FileDebug)))
            {
                var fa = new log4net.Appender.RollingFileAppender()
                {
                    Encoding = Encoding.UTF8,
                    File = init.file.logFolder + String.Format(
                        @"{{0}}-{0}-{1}-ALL.log",
                        init.appName,
                        init.appInstance),
                    LockingModel = new log4net.Appender.FileAppender.MinimalLock(),
                    Threshold = log4net.Core.Level.All,

                    Layout = new log4net.Layout.PatternLayout(pattern)
                    {
                        IgnoresException = true,
                        Header = head,
                        Footer = foot
                    },

                    AppendToFile = true,
                    RollingStyle = log4net.Appender.RollingFileAppender.RollingMode.Date,
                    DatePattern = "yyyyMMdd",
                    StaticLogFileName = false,
                    ImmediateFlush = true
                };
                fa.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(fa);
            }

            if (0 != (init.loggersMask & Logger.FileTrace))
            {
                var fa = new log4net.Appender.RollingFileAppender()
                {
                    Encoding = Encoding.UTF8,
                    File = init.file.logFolder + String.Format(
                        @"{{0}}-{0}-{1}-TRACE.log",
                        init.appName,
                        init.appInstance),
                    LockingModel = new log4net.Appender.FileAppender.MinimalLock(),
                    Threshold = log4net.Core.Level.Trace,

                    Layout = new log4net.Layout.PatternLayout(pattern)
                    {
                        IgnoresException = true,
                        Header = head,
                        Footer = foot
                    },

                    AppendToFile = true,
                    RollingStyle = log4net.Appender.RollingFileAppender.RollingMode.Date,
                    DatePattern = "yyyyMMdd",
                    StaticLogFileName = false,
                    ImmediateFlush = true
                };
                fa.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(fa);
            }

            if (0 != (init.loggersMask & Logger.FileInfo))
            {
                var fa = new log4net.Appender.RollingFileAppender()
                {
                    Encoding = Encoding.UTF8,
                    File = init.file.logFolder + String.Format(@"{{0}}-{0}-{1}-INFO.log", init.appName, init.appInstance),
                    LockingModel = new log4net.Appender.FileAppender.MinimalLock(),
                    Threshold = log4net.Core.Level.Info,

                    Layout = new log4net.Layout.PatternLayout(pattern)
                    {
                        IgnoresException = true,
                        Header = head,
                        Footer = foot
                    },

                    AppendToFile = true,
                    RollingStyle = log4net.Appender.RollingFileAppender.RollingMode.Date,
                    DatePattern = "yyyyMMdd",
                    StaticLogFileName = false,
                    ImmediateFlush = true
                };
                fa.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(fa);
            }

            if (0 != (init.loggersMask & Logger.FileWarnings))
            {
                var fa = new log4net.Appender.RollingFileAppender()
                {
                    Encoding = Encoding.UTF8,
                    File = init.file.logFolder + String.Format(@"{{0}}-{0}-{1}-WARN.log", init.appName, init.appInstance),
                    LockingModel = new log4net.Appender.FileAppender.MinimalLock(),
                    Threshold = log4net.Core.Level.Warn,

                    Layout = new log4net.Layout.PatternLayout(pattern)
                    {
                        IgnoresException = true,
                        Header = head,
                        Footer = foot
                    },

                    AppendToFile = true,
                    RollingStyle = log4net.Appender.RollingFileAppender.RollingMode.Date,
                    DatePattern = "yyyyMMdd",
                    StaticLogFileName = false,
                    ImmediateFlush = true
                };
                fa.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(fa);
            }

            if (0 != (init.loggersMask & Logger.FileErrors))
            {
                var fa = new log4net.Appender.RollingFileAppender()
                {
                    Encoding = Encoding.UTF8,
                    File = init.file.logFolder + String.Format(@"{{0}}-{0}-{1}-ERR.log", init.appName, init.appInstance),
                    LockingModel = new log4net.Appender.FileAppender.MinimalLock(),
                    Threshold = log4net.Core.Level.Error,

                    Layout = new log4net.Layout.PatternLayout(pattern)
                    {
                        IgnoresException = true,
                        Header = head,
                        Footer = foot
                    },

                    AppendToFile = true,
                    RollingStyle = log4net.Appender.RollingFileAppender.RollingMode.Date,
                    DatePattern = "yyyyMMdd",
                    StaticLogFileName = false,
                    ImmediateFlush = true
                };
                fa.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(fa);
            }

            if (0 != (init.loggersMask & Logger.EventErrors))
            {
                var el = new log4net.Appender.EventLogAppender()
                {
                    ApplicationName = init.eventl.application,
                    LogName = init.eventl.log,
                    Threshold = log4net.Core.Level.Error,

                    Layout = new log4net.Layout.PatternLayout(pattern)
                    {
                        IgnoresException = true,
                    }
                };
                el.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(el);
            }

            if (0 != (init.loggersMask & Logger.MailErrors))
            {
                var ml = new log4net.Appender.SmtpAppender()
                {
                    Authentication = log4net.Appender.SmtpAppender.SmtpAuthentication.None,
                    BufferSize = 1,
                    Lossy = false,
                    From = init.mail.from,
                    To = init.mail.to,
                    SmtpHost = init.mail.smtpHost,
                    Port = init.mail.port,
                    Subject = String.Format(
                        "[{0}-{1}] An error has occured",
                        init.appName,
                        init.appInstance),
                    Threshold = log4net.Core.Level.Error,

                    
                    Priority = System.Net.Mail.MailPriority.High,

                    Layout = new log4net.Layout.PatternLayout(pattern)
                    {
                        IgnoresException = false
                    }
                };
                ml.ActivateOptions();
                log4net.Config.BasicConfigurator.Configure(ml);
            }

            {
                CanLogException = false;
                CanLogError = false;
                CanLogWarning = false;
                CanLogInfo = false;
                CanLogTrace = false;
                CanLogDebug = false;

                if (0 != (init.loggersMask & (
                    Logger.MailErrors |
                    Logger.EventErrors |
                    Logger.FileErrors |
                    Logger.ConsoleErrors)))
                {
                    CanLogException = true;
                    CanLogError = true;
                }

                if (0 != (init.loggersMask & (
                    Logger.FileWarnings |
                    Logger.ConsoleWarnings)))
                {
                    CanLogException = true;
                    CanLogError = true;
                    CanLogWarning = true;
                }

                if (0 != (init.loggersMask & (
                    Logger.FileInfo)))
                {
                    CanLogException = true;
                    CanLogError = true;
                    CanLogWarning = true;
                    CanLogInfo = true;
                }

                if (0 != (init.loggersMask & (
                    Logger.FileTrace)))
                {
                    CanLogException = true;
                    CanLogError = true;
                    CanLogWarning = true;
                    CanLogInfo = true;
                    CanLogTrace = true;
                }

                if (0 != (init.loggersMask & (
                    Logger.FileAll | 
                    Logger.FileDebug |
                    Logger.ConsoleAll)))
                {
                    CanLogException = true;
                    CanLogError = true;
                    CanLogWarning = true;
                    CanLogInfo = true;
                    CanLogTrace = true;
                    CanLogDebug = true;
                }
            }

            {
                FileLoggerLevel = TraceLevel.Off;

                if (0 != (init.loggersMask & Logger.FileErrors))
                {
                    FileLoggerLevel = TraceLevel.Error;
                }
                if (0 != (init.loggersMask & Logger.FileWarnings))
                {
                    FileLoggerLevel = TraceLevel.Warning;
                }
                if (0 != (init.loggersMask & Logger.FileInfo))
                {
                    FileLoggerLevel = TraceLevel.Info;
                }
                if (0 != (init.loggersMask & Logger.FileTrace))
                {
                    FileLoggerLevel = TraceLevel.Verbose;
                }
                if ((0 != (init.loggersMask & Logger.FileAll)) ||
                    (0 != (init.loggersMask & Logger.FileDebug)))
                {
                    FileLoggerLevel = TraceLevel.Verbose;
                }
            }

            string reference;
            using (Log.PushRootContext("Logger", out reference))
            {
                if (Log.CanLogTrace) Log.Trace("The logger was initialized: Application='{0}', Instance='{1}', Flavor='{2}'",
                    init.appName,
                    init.appInstance,
                    Settings.Flavor);

                if (!Settings.RevisionIsValid)
                {
                    Log.Error("There a problem when trying to initialize the revision number");
                }
            }
        }

        public static IDisposable PushContext(string s)
        {
            return ThreadContext.Stacks["NDC"].Push(s + " -> ");
        }

        public static IDisposable PushRootContext(string s, out string id)
        {
            id = UniqueId.GetPseudo().ToUpper();
            return PushContext(id + " -> " + s);
        }

        public static void Debug(string message, params object[] args)
        {
            if ((null == args) || (0 == args.Length))
            {
                log.DebugFormat("{0}", message);
            }
            else
            {
                log.DebugFormat(message, args);
            }
        }
        public static void Trace(string message, params object[] args)
        {
            if ((null == args) || (0 == args.Length))
            {
                log.TraceFormat("{0}", message);
            }
            else
            {
                log.TraceFormat(message, args);
            }
        }
        public static void Info(string message, params object[] args)
        {
            if ((null == args) || (0 == args.Length))
            {
                log.Info(message);
            }
            else
            {
                log.InfoFormat(message, args);
            }
        }
        public static void Warning(string message, params object[] args)
        {
            if ((null == args) || (0 == args.Length))
            {
                log.Warn(message);
            }
            else
            {
                log.WarnFormat(message, args);
            }
        }
        public static void Error(string message, params object[] args)
        {
            if ((null == args) || (0 == args.Length))
            {
                log.Error(message);
            }
            else
            {
                log.ErrorFormat(message, args);
            }
        }
        public static void Exception(Exception ex, string message, params object[] args)
        {
            if ((null == args) || (0 == args.Length))
            {
                log.Fatal(message, ex);
            }
            else
            {
                log.Fatal(String.Format(message, args), ex);
            }
        }
        public static void Exception(Exception ex)
        {
            Exception(ex, "An exception has occurred");
        }

        public static bool CanLogException { get; internal set; }
        public static bool CanLogError { get; internal set; }
        public static bool CanLogWarning { get; internal set; }
        public static bool CanLogInfo { get; internal set; }
        public static bool CanLogTrace { get; internal set; }
        public static bool CanLogDebug { get; internal set; }

        private static TraceLevel fileLoggerLevel = TraceLevel.Off;
        public static TraceLevel FileLoggerLevel 
        {
            get
            {
                return fileLoggerLevel;
            }
            internal set
            {
                fileLoggerLevel = value;
            }
        }

        public static string GetFileLoggerName(
            string logName,
            string logFolder,
            string app,
            string instance)
        {
            string level = "";
            switch(FileLoggerLevel)
            {
                case TraceLevel.Error:
                    level = "ERR";
                    break;
                case TraceLevel.Warning:
                    level = "WARN";
                    break;
                case TraceLevel.Info:
                    level = "INFO";
                    break;
                case TraceLevel.Verbose:
                    level = "ALL";
                    break;
                default:
                    level = "OFF";
                    break;
            }

            if (!logFolder.EndsWith(@"\"))
            {
                logFolder += @"\";
            }

            return logFolder + String.Format(
                @"{0}-{1}-{2}-{3}.{4}.log",
                Time.ToCanonicalCondensedDateFileString(Time.UTCNow),
                app,
                instance,
                level,
                logName);
        }
    }
}
