﻿// Bitground Software common helpers library
// http://www.bitground.com/
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace Bitground.Common.Helpers
{
    public class AsyncJobs : IDisposable
    {
        #region Construction

        private AsyncJobs()
        {
        }

        private Thread[] workers;
        private string asyncIdentifier;
        public AsyncJobs(
            int workersNo)
        {
            Stats.WorkersNo = workersNo;

            asyncIdentifier = String.Format("ASYNC {0}", UniqueId.GetPseudo().ToUpper());

            using (Log.PushContext(asyncIdentifier))
            {
                if (Log.CanLogInfo) Log.Info("Creating {0} worker threads...", workersNo);

                workers = new Thread[workersNo];
                for (int i = 0; i < workersNo; i++)
                {
                    workers[i] = CreateWorkerThread();
                }

                if (Log.CanLogTrace) Log.Trace("Starting the threads...");
                for (int i = 0; i < workersNo; i++)
                {
                    workers[i].Start(i);
                }
            }
        }

        #endregion

        #region Worker threads

        private bool isTerminating = false;
        private Thread CreateWorkerThread()
        {
            string workerIdentifier = String.Format("{0} ASYNC WORKER", UniqueId.GetPseudo().ToUpper());

            if (Log.CanLogTrace) Log.Trace("Creating worker with identifier {0}...",
                workerIdentifier);

            return new Thread(delegate(object ob)
            {
                using (Log.PushContext(workerIdentifier))
                {
                    try
                    {
                        if (Log.CanLogTrace) Log.Trace("Worker thread was started");

                        bool shouldContinue;
                        do
                        {
                            shouldContinue = false;

                            lock (Stats.lockLockerWaits)
                            {
                                Stats.LockerWaits++;
                            }
                            // waiting for a job assignment
                            locker.WaitOne();

                            JobProcessorF jobProcessor;
                            object data;
                            GetJob(out data, out jobProcessor);

                            if (null != data)
                            {
                                shouldContinue = true;

                                string jobIdentifier = String.Format("JOB {0}", UniqueId.GetPseudo().ToUpper());
                                using (Log.PushContext(jobIdentifier))
                                {
                                    try
                                    {
                                        if (Log.CanLogTrace) Log.Trace("Job was started");
                                        if (Log.CanLogTrace) Log.Trace("We'll process the following job data: {0}",
                                            Params.GetObjectAsString(data));

                                        jobProcessor(data);
                                        if (Log.CanLogTrace) Log.Trace("Job was finished");
                                    }
                                    catch (Exception ex)
                                    {
                                        lock (Stats.lockJobsCrashed)
                                        {
                                            Stats.JobsCrashed++;
                                        }
                                        Log.Exception(ex, "Job has crashed");
                                        Log.Warning("The job data was: {0}",
                                            Params.GetObjectAsString(data));
                                    }
                                }
                            }
                            else
                            {
                                if (!isTerminating)
                                {
                                    shouldContinue = true;
                                }
                                if (Log.CanLogTrace) Log.Trace("Worker got a null job. shouldContinue='{0}'", shouldContinue);
                            }
                        } while (shouldContinue);

                        if (Log.CanLogTrace) Log.Trace("Worker thread was finished");
                    }
                    catch (Exception ex)
                    {
                        Log.Exception(ex, "The worker thread has crashed");
                    }
                }
            });
        }

        #endregion

        #region Queue job

        private Semaphore locker = new Semaphore(0, Int32.MaxValue);
        private delegate void JobProcessorF(object data);
        public delegate int JobF<T>(T data);
        private void QueueJobCore(object data, JobProcessorF jobProcessor)
        {
            using (Log.PushContext(asyncIdentifier))
            {
                AddJob(data, jobProcessor);
                
                locker.Release();

                lock (Stats.lockLockerReleases)
                {
                    Stats.LockerReleases++;
                }
            }
        }

        public void QueueJob<T>(T data, JobF<T> jobProcessor) where T : class
        {
            QueueJobCore(
                data, 
                delegate(object obj)
                {
                    int ret = jobProcessor(obj as T);

                    if (0 == ret)
                    {
                        lock (Stats.lockJobsCompletedSuccess)
                        {
                            Stats.JobsCompletedSuccess++;
                        }
                    }
                    else
                    {
                        lock (Stats.lockJobsCompletedFailure)
                        {
                            Stats.JobsCompletedFailure++;
                        }
                        Log.Error("Failure while executing a job. The error code was '{0}'. The data to process was: {1}",
                            ret,
                            Params.GetObjectAsString(obj));
                    }
                });
        }

        #endregion

        #region Jobs

        // the number of jobs in the queue above we'll issue a warning
        private const int QueueSizeWarningThreshold = 200;

        private class JobItem
        {
            public object data;
            public JobProcessorF jobProcessor;
        }

        private Queue<object> jobs = new Queue<object>();

        private bool jobsWarned = false;
        private int jobsMax = 0;
        private void AddJob(object data, JobProcessorF jobProcessor)
        {
            lock (jobs)
            {
                if ((null == data) || (null == jobProcessor))
                {
                    Log.Error("Trying to push a null object into the queue. This is not allowed, so we'll not add it");
                }
                else
                {
                    Stats.JobsAssigned++;
                    if (Log.CanLogTrace) Log.Trace("Pushing an object to the queue: data={0}", 
                        Params.GetObjectAsString(data));
                    jobs.Enqueue(new JobItem()
                        {
                            data = data,
                            jobProcessor = jobProcessor
                        });
                    Stats.QueueAdds++;

                    int jobsNo = jobs.Count;
                    Stats.QueueMax = Math.Max(Stats.QueueMax, jobsNo);

                    if (QueueSizeWarningThreshold < jobsNo)
                    {
                        // we have too many jobs
                        if (!jobsWarned)
                        {
                            Log.Warning("There are too many items ({0}) in the jobs queue",
                                jobsNo);
                            jobsMax = 0;
                            jobsWarned = true;
                        }

                        jobsMax = Math.Max(jobsMax, jobsNo);
                    }
                }
            }
        }

        private void GetJob(out object data, out JobProcessorF jobProcessor)
        {
            lock (jobs)
            {
                int jobsNo = jobs.Count;
                if (0 < jobsNo)
                {
                    object ob = jobs.Dequeue();

                    if (null == ob)
                    {
                        Log.Error("We've got an null object from the queue");
                        jobProcessor = null;
                        data = null;
                        Stats.QueueGetsReturnNulls++;
                    }
                    else
                    {
                        if (ob is JobItem)
                        {
                            jobProcessor = ((JobItem)ob).jobProcessor;
                            data = ((JobItem)ob).data;
                            Stats.QueueGetsReturnSuccess++;
                        }
                        else
                        {
                            Log.Error("We've got an object from the queue of an unexpected type");
                            jobProcessor = null;
                            data = null;
                            Stats.QueueGetsReturnFailures++;
                        }
                    }

                    if (jobsNo <= QueueSizeWarningThreshold)
                    {
                        if (jobsWarned)
                        {
                            Log.Warning("There were too many items (maximum of {0}) in the jobs queue, but now the load decreased ({1} jobs in the queue at the moment)",
                                jobsMax,
                                jobsNo);
                            jobsWarned = false;
                            jobsMax = 0;
                        }
                    }

                }
                else
                {
                    // the queue is empty
                    jobProcessor = null;
                    data = null;
                    Stats.QueueGetsEmpty++;
                }
            }
        }

        #endregion

        #region Statistics

        private Statistics Stats = new Statistics();
        private class Statistics
        {
            public int WorkersNo = 0;

            public int QueueAdds = 0;
            public int QueueMax = 0;
            public int QueueGetsReturnNulls = 0;
            public int QueueGetsReturnSuccess = 0;
            public int QueueGetsReturnFailures = 0;
            public int QueueGetsEmpty = 0;

            internal object lockLockerWaits = new object();
            public int LockerWaits = 0;

            internal object lockLockerReleases = new object();
            public int LockerReleases = 0;

            public int JobsAssigned = 0;
            
            internal object lockJobsCompletedSuccess = new object();
            public int JobsCompletedSuccess = 0;

            internal object lockJobsCompletedFailure = new object();
            public int JobsCompletedFailure = 0;

            internal object lockJobsCrashed = new object();
            public int JobsCrashed = 0;

            public override string ToString()
            {
                return String.Format("Workers:[No='{0}']; Jobs:[Assigned='{1}', CompletedSuccess='{2}', CompletedFailure='{3}', Crashed='{4}']; Queue:[Adds='{5}', Max='{6}', GetsReturnNulls='{7}', GetsReturnSuccess='{8}', GetsReturnFailures='{9}', GetsEmpty='{10}']; Locker:[Releases='{11}', Waits='{12}']",
                    WorkersNo,
                    JobsAssigned,
                    JobsCompletedSuccess,
                    JobsCompletedFailure,
                    JobsCrashed,
                    QueueAdds,
                    QueueMax,
                    QueueGetsReturnNulls,
                    QueueGetsReturnSuccess,
                    QueueGetsReturnFailures,
                    QueueGetsEmpty,
                    LockerReleases,
                    LockerWaits);
            }
        }

        #endregion

        #region Destruction

        private void Terminate()
        {
            using (Log.PushContext(asyncIdentifier))
            {
                if (Log.CanLogTrace) Log.Trace("Destructing the AsyncJobs object");
                isTerminating = true;
                locker.Release(workers.Length);
                lock (Stats.lockLockerReleases)
                {
                    Stats.LockerReleases += workers.Length;
                }

                if (Log.CanLogTrace) Log.Trace("Waiting for the threads to finish...");
                foreach (Thread th in workers)
                {
                    th.Join();
                }

                if (Log.CanLogTrace) Log.Trace("The threads were finished...");
                if (Log.CanLogInfo) Log.Info("The statistics are: {0}", Stats.ToString());
                if (Log.CanLogInfo) Log.Info("AsyncJobs object terminated");
            }
        }

        void IDisposable.Dispose()
        {
            Terminate();
        }

        public void Dispose()
        {
            Terminate();
        }

        #endregion
    }
}
