﻿// Bitground Software common helpers library
// http://www.bitground.com/
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Configuration;

namespace Bitground.Common.Helpers
{
    public abstract class Settings
    {
        static Settings()
        {
            InitRevision();
        }

        protected static string BaseUrl(string url)
        {
            return url.ToLower().TrimEnd('/');
        }

        protected static string BaseUrlSecure(string url)
        {
            url = BaseUrl(url);
            if (url.StartsWith("http://"))
            {
                url = "https://" + url.Substring("http://".Length);
            }
            else
            {
                Log.Error("BaseUrlSecure got an invalid url ({0})", url);
            }
            return url;
        }

        public static bool DEBUG
        {
            get
            {
#if DEBUG
                return true;
#else
                return false;
#endif
            }
        }

        public static bool RELEASE
        {
            get
            {
#if DEBUG
                return false;
#else
                return true;
#endif
            }
        }

        public static string Flavor
        {
            get
            {
                if (Settings.RELEASE)
                {
                    return "RELEASE";
                }
                return "DEBUG";
            }

        }

        #region Revision

        private static string revision = null;
        private static string revisionNull = "000UNK";
        public static string Revision
        {
            get
            {
                return revision;
            }
        }

        internal static bool RevisionIsValid
        {
            get
            {
                return (revision != revisionNull);
            }
        }

        private static void InitRevision()
        {
            string v = "";
            try
            {
                v = Bitground.Common.Helpers.Revision.getMaxVersion();
                v = v.TrimEnd('M', 'P', 'S');

                v = Strings.CutTo(
                    6,
                    Strings.PaddingBeginning(
                        v,
                        6,
                        '0'));

                revision = v;
            }
            catch (Exception)
            {
                // swallow the exception

                revision = revisionNull;
            }
        }

        #endregion

        #region Copyright

        private static string CopyrightYear
        {
            get
            {
                /* copyright as period
                int startYear = 2009;
                string ret;

                int year = Time.UTCNow.Year;

                if (startYear < year)
                {
                    ret = String.Format(
                        "{0}-{1}",
                        Localization.ToInvariantString(startYear),
                        Localization.ToInvariantString(year));
                }
                else
                {
                    ret = Localization.ToInvariantString(startYear);
                }

                return ret;
                */

                // copyright as the current year
                return Localization.ToInvariantString(Time.UTCNow.Year);
            }
        }

        public static string CopyrightPrefix
        {
            get
            {
                return String.Format(
                    "&copy; {0}",
                    CopyrightYear);
            }
        }

        #endregion

        #region Build date

        private static DateTime? buildDate = null;
        private static object buildDateLocker = new object();
        public static DateTime? BuildDate
        {
            get
            {
                if (null == buildDate)
                {
                    lock (buildDateLocker)
                    {
                        if (null == buildDate)
                        {
                            try
                            {
                                string fileName = Assembly.GetExecutingAssembly().Location;
                                DateTime assemblyCreationDate = System.IO.File.GetLastWriteTime(fileName);
                                DateTime bd = assemblyCreationDate.ToUniversalTime();
                                if (bd == new DateTime(1601, 1, 1))
                                {
                                    Log.Error("Failure while trying to get the build date. Perhaps the file '{0}' was not found", fileName);
                                    bd = DateTime.MinValue;
                                }
                                buildDate = bd;
                            }
                            catch (Exception ex)
                            {
                                Log.Exception(ex, "Failure while trying to get the build date");
                                buildDate = DateTime.MinValue;
                            }
                        }
                    }
                }

                if (DateTime.MinValue == buildDate)
                {
                    return null;
                }

                return buildDate;
            }
        }

        #endregion

        #region Core config functions

        protected static string GetString(string key)
        {
            string val = ConfigurationManager.AppSettings[key];

            Debug.Assert(null != val, "Could not find the configuration key '{0}'", key);

            return val;
        }
        protected static double GetDouble(string key)
        {
            return Double.Parse(GetString(key));
        }
        protected static int GetInt(string key)
        {
            return Int32.Parse(GetString(key));
        }
        protected static bool GetBool(string key)
        {
            return Boolean.Parse(GetString(key));
        }
        protected static E GetEnum<E>(string key)
        {
            return (E)Enum.Parse(typeof(E), GetString(key));
        }

        #endregion
    }
}
