﻿// Bitground Software common helpers library
// http://www.bitground.com/
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Threading;
using System.Net;
using System.IO;

namespace Bitground.Common.Helpers
{
    public abstract class TrueRandom
    {
        internal class Stats
        {
            // the total count of numbers requested
            internal int TotalRequests { get; set; }

            // the total ount of numbers returned (should be equal to TotalRequests)
            internal int TotalGet { get; set; }

            // the total numbers lost
            internal int TotalLost { get; set; }

            // the total numbers requested from real random numbers generator
            internal int TotalGrab { get; set; }

            // the total bits requested
            internal UInt64 TotalBits { get; set; }

            // the total chunks of 100k bits
            internal UInt64 TotalBits100kChunks { get; set; }

            // the total number of pseudo random numbers generated
            internal int TotalPseudo { get; set; }

            // the last queue size
            internal int LastQueueSize { get; set; }

            public void Check()
            {
                using (Log.PushContext("Check"))
                {
                    if (Log.CanLogTrace)
                    {
                        Log.Trace("TotalRequests = {0}", TotalRequests);
                        Log.Trace("TotalGet = {0}", TotalGet);
                        Log.Trace("TotalGrab = {0}", TotalGrab);
                        Log.Trace("TotalBits = {0}", TotalBits);
                        Log.Trace("TotalLost = {0}", TotalLost);
                        Log.Trace("TotalPseudo = {0}", TotalPseudo);
                        Log.Trace("LastQueueSize = {0}", LastQueueSize);

                        Log.Trace("TotalRequests + LastQueueSize = TotalGrab - TotalLost + TotalPseudo --> " +
                                  "{0} + {1} = {2} - {3} + {4} --> " +
                                  "{5} = {6} ?",
                                TotalRequests,
                                LastQueueSize,
                                TotalGrab,
                                TotalLost,
                                TotalPseudo,
                                TotalRequests + LastQueueSize,
                                TotalGrab - TotalLost + TotalPseudo);
                    }

                    if (TotalRequests + LastQueueSize == TotalGrab - TotalLost + TotalPseudo)
                    {
                        if (Log.CanLogInfo) Log.Info("The Check worked!");
                    }
                    else
                    {
                        Log.Error("The Check did NOT work!");
                    }
                }
            }
        }

        private static TrueRandomGenerator trng = null;

        private static bool m_bInitialized = false;
        public static void Initialize()
        {
            if (!m_bInitialized)
            {
                lock (typeof(TrueRandom))
                {
                    if (!m_bInitialized)
                    {
                        trng = new TrueRandomGenerator();
                        trng.Initialize();
                        m_bInitialized = true;
                    }
                }
            }
        }

        public static UInt32[] GetUInt32Array(int count)
        {
            lock (typeof(TrueRandom))
            {
                if (!m_bInitialized)
                {
                    throw new CException("TrueRandom was not initialized");
                }
            }

            if (count <= 0)
            {
                return new UInt32[0];
            }

            TrueRandomGenerator.TotalRequests += count;

#if DEBUG
            if (null == trng)
            {
                throw new CException("TrueRandom was not initialized");
            }
#endif

            UInt32[] ar = trng.Get(count);
            Debug.Assert(count == ar.Length, "Failure in GetUInt32Array ({0} != {1})", count, ar.Length);
            
            return ar;
        }

        public static UInt32 GetUInt32()
        {
            UInt32[] ar = GetUInt32Array(1);
            return ar[0];
        }

        public static UInt32 GetUInt32(UInt32 max)
        {
            // we'll generate an integer in the [0 .. max) interval (max should be included)

            UInt32 ui;
            using (Log.PushContext("GetUInt32"))
            {
                // if smaller or equal than 1, then return 0
                if (max <= 1)
                {
                    return 0;
                }
                // so max is 2 or more


                // get 32 true random bits
                ui = GetUInt32();

                ui = PseudoRandom.ScaleRandom(ui, max);

                Debug.Assert((0 <= ui) && (ui < max), "Failure in GetUInt32 (ui={0}, max={1})", ui, max);
            }

            return ui;
        }
        public static Int32 GetInt32(Int32 max)
        {
            // if smaller or equal than 1, then return 0
            if (max <= 1)
            {
                return 0;
            }

            return (Int32)GetUInt32((UInt32)max);
        }

        public static byte[] GetByteArray(int count)
        {
            if (count <= 0)
            {
                return new byte[0];
            }

            int c = (count + 3) / 4;
            UInt32[] ui = GetUInt32Array(c);

            byte[] ar = new byte[count];

            int p = 0;
            for (int i = 0; i < c - 1; i++)
            {
                UInt32 u = ui[i];
                ar[p++] = u.GetByte(0);
                ar[p++] = u.GetByte(1);
                ar[p++] = u.GetByte(2);
                ar[p++] = u.GetByte(3);
            }

            {
                UInt32 last = ui[c - 1];
                for (int i = 0; i < 1 + (count - 1) % 4; i++)
                {
                    ar[p++] = last.GetByte(i);
                }
            }

            return ar;
        }

        internal static Stats Statistics
        {
            get
            {
                return new Stats()
                {
                    TotalRequests = TrueRandomGenerator.TotalRequests,
                    TotalGet = TrueRandomGenerator.TotalGet,
                    TotalLost = TrueRandomGenerator.TotalLost,
                    TotalGrab = TrueRandomGenerator.TotalGrab,
                    TotalBits = TrueRandomGenerator.TotalBits,
                    TotalBits100kChunks = TrueRandomGenerator.TotalBits100kChunks,
                    TotalPseudo = TrueRandomGenerator.TotalPseudo,
                    LastQueueSize = TrueRandomGenerator.LastQueueSize
                };
            }
        }

        private static bool m_bDisposed = false;
        public static void Dispose()
        {
            try
            {
                if ((m_bInitialized) && (!m_bDisposed))
                {
                    lock (typeof(TrueRandom))
                    {
                        if ((m_bInitialized) && (!m_bDisposed))
                        {
                            trng.Dispose();
                            if (Log.CanLogInfo) Log.Info("The true randoms object was disposed");

                            Statistics.Check();
                            m_bDisposed = true;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Log.Exception(ex, "Failure while trying to dispose the TrueRandom generator");
            }
        }
    }

}
