﻿// Bitground Software common helpers library
// http://www.bitground.com/
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Threading;
using System.Net;
using System.IO;

namespace Bitground.Common.Helpers
{
    internal class TrueRandomGenerator : IDisposable
    {
        internal Int64 GetQuota()
        {
            Int64 quota = -1;
            using (Log.PushContext("GetQuota"))
            {
                if (m_bUseTRNG)
                {
                    // getting the quota of the TRNG
                    string quotaURL = "https://www.random.org/quota/?format=plain";

                    if (Log.CanLogTrace) Log.Trace("Requesting quota ({0})...", quotaURL);

                    string s;

                    try
                    {
                        WebRequest req = (WebRequest)WebRequest.Create(quotaURL);
                        StreamReader sr = new StreamReader(req.GetResponse().GetResponseStream());
                        s = sr.ReadToEnd();

                        Int64.TryParse(s, out quota);
                    }
                    catch (Exception ex)
                    {
                        Log.Exception(ex, "Failure in GetQuota");
                    }
                }
                else
                {
                    // when using PRNG, the quota is infinite
                    quota = Int64.MaxValue;
                }

                if (Log.CanLogTrace) Log.Trace("The quota is {0}", quota);
                if ((0 <= quota) && (quota < 100000))
                {
                    Log.Error("The quota is too low ({0})", quota);
                }
            }
            return quota;
        }

        // the random pool size
        // m_poolSize - the equivalent of 10 strings of 2048bits each
        //private const int m_poolSize = (10 * 2048) / 16;
        private const int m_poolSize = 1500;

        // the size under which the buffer should be filled in with true random numbers
        // m_thresholdTrueRandom - the equivalent of 5 strings of 2048bits each
        //private const int m_thresholdTrueRandom = (5 * 2048) / 16;
        private const int m_thresholdTrueRandom = 200;

        // the size under which the buffer should will be filled in with pseudo random numbers
        // should be greater than the maximum value we will get in the Get method
        // m_thresholdPseudoRandom - the equivalent of one 2048bits string (+1)
        // that's it, we cannot request more than 2048 random bits at one call
        //private const int m_thresholdPseudoRandom = 2048 / 16 + 1;
        private const int m_thresholdPseudoRandom = 50;

        // the random numbers pool
        private List<UInt32> m_pool = new List<UInt32>();

        // the thread responsible for fetching true randoms
        private Thread m_thGetTrueRandoms = null;

        // true if we should indeed use the TRNG
        // if too many requests are made to the TRNG we might disable it and fall back to PRNG (pseudo-RNG)
        private bool m_bUseTRNG = true;

        public TrueRandomGenerator()
        {
        }

        internal void Initialize()
        {
            if (Log.CanLogInfo) Log.Info("Initializing the TrueRandom class");
            Debug.Assert(0 < m_thresholdPseudoRandom);
            Debug.Assert(m_thresholdPseudoRandom < m_thresholdTrueRandom);
            Debug.Assert(m_thresholdTrueRandom < m_poolSize);

            if (!m_bUseTRNG)
            {
                Log.Warning("The true random generator is generating pseudo-randoms in fact");
            }

            if (Log.CanLogInfo) Log.Info("Spawning the get true randoms thread");
            m_thGetTrueRandoms = new Thread(this.GetTrueRandoms);
            m_thGetTrueRandoms.Start();
        }

        private bool shouldExit = false;

        public void Dispose()
        {
            if (Log.CanLogInfo) Log.Info("Trying to abort the get true randoms thread");

            shouldExit = true;
            //m_thGetTrueRandoms.Abort();
            ev.Set();
            m_thGetTrueRandoms.Join();

            lock (m_pool)
            {
                TrueRandomGenerator.LastQueueSize = m_pool.Count;
            }

            m_thGetTrueRandoms = null;

            ev.Close();
            ev = null;
        }

        public UInt32[] Get(int count)
        {
            // return an array of count random values

            Debug.Assert(count < m_thresholdPseudoRandom, "count should be smaller than m_thresholdPseudoRandom");
            UInt32[] ar = new UInt32[count];

            using (Log.PushContext("[TRNG] Get"))
            {
                lock (m_pool)
                {
                    CheckPool();
                    Debug.Assert(count < m_pool.Count);

                    for (int i = 0; i < count; i++)
                    {
                        ar[i] = m_pool[0];
                        m_pool.RemoveAt(0);
                    }

                    if (Log.CanLogDebug) Log.Debug("[GET] we returned {0} numbers", count);
                    TrueRandomGenerator.TotalGet += count;
                    if (Log.CanLogDebug) Log.Debug("pool size: " + m_pool.Count);
                }
            }

            return ar;
        }

        private void CheckPool()
        {
            // check to see if we need to fill the pool
            // this function should ensure that there are at least m_thresholdPseudoRandom numbers in the pool

            // when calling this, the pool should be locked

            using (Log.PushContext("CheckPool"))
            {
                if (Log.CanLogDebug) Log.Debug("The pool is: {0}", ToString());

                int l = m_pool.Count;
                if (l < m_thresholdPseudoRandom)
                {
                    // there are too few numbers, perhaps the true random generator didn't work
                    // we'll fill the pool with pseudo random numbers
                    if (Log.CanLogInfo) Log.Info("There are too few numbers (l={0}, m_thresholdPseudoRandom={1}), perhaps the true random generator didn't work. We'll fill the pool with pseudo random numbers.", l, m_thresholdPseudoRandom);

                    for (int i = 0; i < m_poolSize - l; i++)
                    {
                        m_pool.Add(PseudoRandom.GetUInt32());
                    }
                    if (Log.CanLogInfo) Log.Info("[PSEUDO] we got {0} pseudo numbers", m_poolSize - l);
                    TrueRandomGenerator.TotalPseudo += m_poolSize - l;
                    if (Log.CanLogDebug) Log.Debug("The pool is (a): {0}", ToString());
                }
                else
                {
                    if (l < m_thresholdTrueRandom)
                    {
                        // we are under the true random threshold, so we should start getting new random numbers
                        if (Log.CanLogInfo) Log.Info("We are under the m_thresholdTrueRandom threshold ({0} < {1})", l, m_thresholdTrueRandom);

                        // unlock here the thread
                        ev.Set();
                    }
                    else
                    {
                        // we don't do anything, there are enough random numbers in the pool
                        if (Log.CanLogDebug) Log.Debug("We don't do anything, there are enough random numbers in the pool");
                    }
                }
            }
        }


        private EventWaitHandle ev = new EventWaitHandle(true, EventResetMode.AutoReset); // not signaled
        private void GetTrueRandoms()
        {
            string referenceId;
            using (Log.PushRootContext("[TRNG] GetTrueRandoms", out referenceId))
            {
                if (Log.CanLogInfo) Log.Info("Starting...");
                if (Log.CanLogInfo) Log.Info("The quota at startup is {0}", GetQuota());

                while (!shouldExit)
                {
                    ev.WaitOne(); // waiting for a signal

                    if (!shouldExit)
                    {
                        if (Log.CanLogInfo) Log.Info("Start getting true random numbers...");

                        int count = 0;
                        lock (m_pool)
                        {
                            count = m_poolSize - m_pool.Count;
                        }

                        if (0 < count)
                        {
                            if (Log.CanLogInfo) Log.Info("[GRAB] we'll grab {0} numbers", count);
                            TrueRandomGenerator.TotalGrab += count;


                            List<UInt32> ret = new List<UInt32>();
                            using (Log.PushContext("WS call"))
                            {
                                // we'll try to get count true randoms numbers
                                string randomURL = String.Format(
                                    "https://www.random.org/integers/?num={0}&min=0&max=65535&col=1&base=10&format=plain&rnd=new",
                                    count * 2); // we need count * 2 because we are getting UInt16 from the service
                                TrueRandomGenerator.TotalBits += (UInt64)((count * 2) * 16);

                                string s = "";

                                if (m_bUseTRNG)
                                {
                                    if (Log.CanLogInfo) Log.Info(
                                        "Calling the webservice ({0})",
                                        randomURL);
                                    try
                                    {
                                        WebRequest req = (WebRequest)WebRequest.Create(randomURL);
                                        StreamReader sr = new StreamReader(req.GetResponse().GetResponseStream());
                                        s = sr.ReadToEnd();
                                    }
                                    catch (Exception ex)
                                    {
                                        Log.Exception(ex, "Failure when calling the webservice");
                                    }

                                    if (TrueRandomGenerator.TotalBits100kChunks < TrueRandomGenerator.TotalBits / 100000)
                                    {
                                        TrueRandomGenerator.TotalBits100kChunks = TrueRandomGenerator.TotalBits / 100000;
                                        if (Log.CanLogInfo) Log.Info(
                                            "At least another 100k bits passed. The quota is now: {0}",
                                            GetQuota());
                                    }
                                }
                                else
                                {
                                    for (int i = 0; i < count * 2; i++)
                                    {
                                        s += PseudoRandom.GetUInt16().ToString() + "\n";
                                    }
                                }

                                string[] arStringValues = s.Split('\n', '\r');

                                int l = Math.Min(count * 2, arStringValues.Length);
                                List<UInt16> arValues = new List<UInt16>();

                                for (int i = 0; i < l; i++)
                                {
                                    UInt16 ui;
                                    if (UInt16.TryParse(arStringValues[i].Trim(), out ui))
                                    {
                                        arValues.Add(ui);
                                    }
                                }

                                for (int i = 0; i < arValues.Count - 1; i += 2)
                                {
                                    // get one UInt32 number
                                    UInt32 ui = new UInt32().FromUnsignedShort(arValues[i], arValues[i + 1]);

                                    // XOR it with a pseudo number
                                    // this will assure us that if someone is faking the true random service, we'll end up at least having pseudo-random numbers
                                    ui ^= PseudoRandom.GetUInt32();

                                    ret.Add(ui);
                                }


                                int rt = ret.Count;
                                if (rt < count)
                                {
                                    Log.Warning("For some reason we didn't get all the numbers we requested. (we've got only {0} instead of {1}). We'll fill the rest of the array with pseudonumbers'", rt, count);
                                    for (int i = 0; i < count - rt; i++)
                                    {
                                        ret.Add(PseudoRandom.GetUInt32());
                                    }
                                }

                                if (Log.CanLogInfo) Log.Info("We called the webservice; we've got {0} numbers", ret.Count);
                            }
                            Debug.Assert(ret.Count == count);

                            // now putting the numbers in the pool
                            lock (m_pool)
                            {
                                using (Log.PushContext("Adding to the pool"))
                                {
                                    if (Log.CanLogTrace) Log.Trace("The pool is (b): {0}", ToString());
                                    int cnt = Math.Min(ret.Count, m_poolSize - m_pool.Count);
                                    for (int i = 0; i < cnt; i++)
                                    {
                                        m_pool.Add(ret[i]);
                                    }
                                    if (0 < count - cnt)
                                    {
                                        if (Log.CanLogInfo) Log.Info("[LOST] we've lost {0} numbers", count - cnt);
                                    }
                                    TrueRandomGenerator.TotalLost += count - cnt;
                                    if (Log.CanLogInfo) Log.Info("We've just added {0} numbers to the pool", count - cnt);
                                    if (Log.CanLogTrace) Log.Trace("The pool is (c): {0}", ToString());
                                }
                            }
                        }
                        else
                        {
                            if (Log.CanLogInfo) Log.Info("GetTrueRandoms was called, but there are no numbers to grab");
                        }
                    }
                }
            }
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();

            lock (m_pool)
            {
                sb.AppendFormat("({0}): ", m_pool.Count);
                /*
                foreach (UInt32 i in m_pool)
                {
                    sb.AppendFormat(
                        "{0} ",
                        i);
                }*/
                const int maxLengthDisplay = 10;
                for (int i = 0; i < Math.Min(maxLengthDisplay, m_pool.Count); i++)
                {
                    sb.AppendFormat(
                        "{0} ",
                        m_pool[i]);
                }
                if (maxLengthDisplay < m_pool.Count)
                {
                    sb.Append("...");
                }
            }

            return sb.ToString();
        }

        public static int TotalRequests = 0;
        public static int TotalGet = 0;
        public static int TotalLost = 0;
        public static int TotalGrab = 0;
        public static UInt64 TotalBits = 0;
        public static UInt64 TotalBits100kChunks = 0;
        public static int TotalPseudo = 0;
        public static int LastQueueSize = 0;
    }
}
