// Bitground Software common helpers library
// http://www.bitground.com/
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;
using System.Data;
using System.IO;
using System.Web;

namespace Bitground.Common.Helpers
{
    public abstract class Localization
    {
        #region Constants
        
        public const string KEYPREFIX = "KEY:";
        
        #endregion

        #region Initialization

        public static CultureInfo InvariantCulture;
        public static DateTimeFormatInfo InvariantDateTimeFormat;
        public static NumberFormatInfo InvariantNumberFormat;
        static Localization()
        {
            InvariantCulture = CultureInfo.InvariantCulture;
            InvariantDateTimeFormat = InvariantCulture.DateTimeFormat;
            InvariantNumberFormat = InvariantCulture.NumberFormat;
        }

        #endregion

        #region Invariant conversions from string to type (Parse)

        public static bool ToInvariantBoolean(string strValue)
        {
            return Boolean.Parse(strValue);
        }
        public static short ToInvariantInt16(string strValue)
        {
            return Int16.Parse(strValue, InvariantNumberFormat);
        }
        public static int ToInvariantInt32(string strValue)
        {
            return Int32.Parse(strValue, InvariantNumberFormat);
        }
        public static long ToInvariantInt64(string strValue)
        {
            return Int64.Parse(strValue, InvariantNumberFormat);
        }
        public static float ToInvariantSingle(string strValue)
        {
            return Single.Parse(strValue, InvariantNumberFormat);
        }
        public static double ToInvariantDouble(string strValue)
        {
            return Double.Parse(strValue, InvariantNumberFormat);
        }
        public static DateTime ToInvariantDateTime(string strValue)
        {
            return DateTime.Parse(strValue, InvariantDateTimeFormat);
        }

        #endregion

        #region Invariant conversions from object to type (Convert)

        public static bool ConvertToInvariantBoolean(object obValue)
        {
            return Convert.ToBoolean(obValue, InvariantNumberFormat);
        }
        public static short ConvertToInvariantInt16(object obValue)
        {
            return Convert.ToInt16(obValue, InvariantNumberFormat);
        }
        public static int ConvertToInvariantInt32(object obValue)
        {
            return Convert.ToInt32(obValue, InvariantNumberFormat);
        }
        public static long ConvertToInvariantInt64(object obValue)
        {
            return Convert.ToInt64(obValue, InvariantNumberFormat);
        }
        public static float ConvertToInvariantSingle(object obValue)
        {
            return Convert.ToSingle(obValue, InvariantNumberFormat);
        }
        public static double ConvertToInvariantDouble(object obValue)
        {
            return Convert.ToDouble(obValue, InvariantNumberFormat);
        }
        public static DateTime ConvertToInvariantDateTime(object obValue)
        {
            return Convert.ToDateTime(obValue, InvariantDateTimeFormat);
        }
        public static string ConvertToInvariantString(object obValue)
        {
            if (obValue is DateTime)
            {
                return Convert.ToString(obValue, InvariantDateTimeFormat);
            }
            return Convert.ToString(obValue, InvariantNumberFormat);
        }

        #endregion

        #region Culture dependent conversions from string to type (Parse)

        public static bool ToBoolean(string strValue)
        {
            return Boolean.Parse(strValue);
        }
        public static short ToInt16(string strValue)
        {
            return Int16.Parse(strValue);
        }
        public static int ToInt32(string strValue)
        {
            return Int32.Parse(strValue);
        }
        public static long ToInt64(string strValue)
        {
            return Int64.Parse(strValue);
        }
        public static float ToSingle(string strValue)
        {
            return Single.Parse(strValue);
        }
        public static double ToDouble(string strValue)
        {
            return Double.Parse(strValue);
        }
        public static DateTime ToDateTime(string strValue)
        {
            return DateTime.Parse(strValue);
        }

        #endregion

        #region Culture dependent conversions from object to type (Convert)

        public static bool ConvertToBoolean(object obValue)
        {
            return Convert.ToBoolean(obValue);
        }
        public static short ConvertToInt16(object obValue)
        {
            return Convert.ToInt16(obValue);
        }
        public static int ConvertToInt32(object obValue)
        {
            return Convert.ToInt32(obValue);
        }
        public static long ConvertToInt64(object obValue)
        {
            return Convert.ToInt64(obValue);
        }
        public static float ConvertToSingle(object obValue)
        {
            return Convert.ToSingle(obValue);
        }
        public static double ConvertToDouble(object obValue)
        {
            return Convert.ToDouble(obValue);
        }
        public static DateTime ConvertToDateTime(object obValue)
        {
            return Convert.ToDateTime(obValue);
        }

        #endregion

        #region Invariant conversions from type to string

        public static string ToInvariantString(bool value)
        {
            return value.ToString(InvariantNumberFormat);
        }
        public static string ToInvariantString(short value)
        {
            return value.ToString(InvariantNumberFormat);
        }
        public static string ToInvariantString(int value)
        {
            return value.ToString(InvariantNumberFormat);
        }
        public static string ToInvariantString(long value)
        {
            return value.ToString(InvariantNumberFormat);
        }
        public static string ToInvariantString(float value)
        {
            return value.ToString(InvariantNumberFormat);
        }
        public static string ToInvariantString(double value)
        {
            return value.ToString(InvariantNumberFormat);
        }
        public static string ToInvariantString(DateTime value)
        {
            return value.ToString(InvariantDateTimeFormat);
        }

        public static string ToInvariantString(object obValue)
        {
            if (obValue is int)
            {
                return ToInvariantString((int)obValue);
            }
            if (obValue is string)
            {
                return ToInvariantString((string)obValue);
            }
            if (obValue is bool)
            {
                return ToInvariantString((bool)obValue);
            }
            if (obValue is DateTime)
            {
                return ToInvariantString((DateTime)obValue);
            }
            if (obValue is long)
            {
                return ToInvariantString((long)obValue);
            }
            if (obValue is double)
            {
                return ToInvariantString((double)obValue);
            }
            if (obValue is short)
            {
                return ToInvariantString((short)obValue);
            }
            if (obValue is float)
            {
                return ToInvariantString((float)obValue);
            }

            throw new CException(
                "Unhandled data type '{0}' in Localization.ToInvariantString",
                obValue.GetType());
        }

        public static string ToInvariantString(string value)
        {
            return value;
        }

        #endregion

        #region Invariant conversions with format from type to string

        public static string ToInvariantString(string format, short value)
        {
            return value.ToString(format, InvariantNumberFormat);
        }
        public static string ToInvariantString(string format, int value)
        {
            return value.ToString(format, InvariantNumberFormat);
        }
        public static string ToInvariantString(string format, long value)
        {
            return value.ToString(format, InvariantNumberFormat);
        }
        public static string ToInvariantString(string format, float value)
        {
            return value.ToString(format, InvariantNumberFormat);
        }
        public static string ToInvariantString(string format, double value)
        {
            return value.ToString(format, InvariantNumberFormat);
        }
        public static string ToInvariantString(string format, DateTime value)
        {
            return value.ToString(format, InvariantDateTimeFormat);
        }

        #endregion

        #region Culture dependent conversions from type to string

        public static string ToString(bool value)
        {
            return value.ToString();
        }
        public static string ToString(short value)
        {
            return value.ToString();
        }
        public static string ToString(int value)
        {
            return value.ToString();
        }
        public static string ToString(long value)
        {
            return value.ToString();
        }
        public static string ToString(float value)
        {
            return value.ToString();
        }
        public static string ToString(double value)
        {
            return value.ToString();
        }
        public static string ToString(DateTime value)
        {
            return value.ToString();
        }
        public static string ToString(object value)
        {
            return value.ToString();
        }

        #endregion

        #region Culture dependent conversions with format from type to string

        public static string ToString(string format, short value)
        {
            return value.ToString(format);
        }
        public static string ToString(string format, int value)
        {
            return value.ToString(format);
        }
        public static string ToString(string format, long value)
        {
            return value.ToString(format);
        }
        public static string ToString(string format, float value)
        {
            return value.ToString(format);
        }
        public static string ToString(string format, double value)
        {
            return value.ToString(format);
        }
        public static string ToString(string format, DateTime value)
        {
            return value.ToString(format);
        }

        #endregion

        #region Invariant conversions from object to string

        public static string ObjectBooleanToInvariantString(object value)
        {
            return ToInvariantString(ConvertToInvariantBoolean(value));
        }
        public static string ObjectInt16ToInvariantString(object value)
        {
            return ToInvariantString(ConvertToInvariantInt16(value));
        }
        public static string ObjectInt32ToInvariantString(object value)
        {
            return ToInvariantString(ConvertToInvariantInt32(value));
        }
        public static string ObjectInt64ToInvariantString(object value)
        {
            return ToInvariantString(ConvertToInvariantInt64(value));
        }
        public static string ObjectSingleToInvariantString(object value)
        {
            return ToInvariantString(ConvertToInvariantSingle(value));
        }
        public static string ObjectDoubleToInvariantString(object value)
        {
            return ToInvariantString(ConvertToInvariantDouble(value));
        }
        public static string ObjectDateTimeToInvariantString(object value)
        {
            return ToInvariantString(ConvertToInvariantDateTime(value));
        }

        #endregion

        #region Culture dependent conversions from object to string

        public static string ObjectBooleanToString(object value)
        {
            return ToString(ConvertToBoolean(value));
        }
        public static string ObjectInt16ToString(object value)
        {
            return ToString(ConvertToInt16(value));
        }
        public static string ObjectInt32ToString(object value)
        {
            return ToString(ConvertToInt32(value));
        }
        public static string ObjectInt64ToString(object value)
        {
            return ToString(ConvertToInt64(value));
        }
        public static string ObjectSingleToString(object value)
        {
            return ToString(ConvertToSingle(value));
        }
        public static string ObjectDoubleToString(object value)
        {
            return ToString(ConvertToDouble(value));
        }
        public static string ObjectDateTimeToString(object value)
        {
            return ToString(ConvertToDateTime(value));
        }

        #endregion

        #region Culture dependent conversion from object to type with default value

        public static bool DefaultToIfNullOrEmpty(object obValue, bool bDefaultValue)
        {
            return (null == obValue) ? bDefaultValue : DefaultToIfNullOrEmpty(ToString(obValue), bDefaultValue);
        }
        public static int DefaultToIfNullOrEmpty(object obValue, int intDefaultValue)
        {
            return (null == obValue) ? intDefaultValue : DefaultToIfNullOrEmpty(ToString(obValue), intDefaultValue);
        }
        public static long DefaultToIfNullOrEmpty(object obValue, long longDefaultValue)
        {
            return (null == obValue) ? longDefaultValue : DefaultToIfNullOrEmpty(ToString(obValue), longDefaultValue);
        }
        public static double DefaultToIfNullOrEmpty(object obValue, double dblDefaultValue)
        {
            return (null == obValue) ? dblDefaultValue : DefaultToIfNullOrEmpty(ToString(obValue), dblDefaultValue);
        }
        public static string DefaultToIfNullOrEmpty(object obValue, string strDefaultValue)
        {
            return (null == obValue) ? strDefaultValue : DefaultToIfNullOrEmpty(ToString(obValue), strDefaultValue);
        }
        public static DateTime DefaultToIfNullOrEmpty(object obValue, DateTime dtDefaultValue)
        {
            return (null == obValue) ? dtDefaultValue : DefaultToIfNullOrEmpty(ToString(obValue), dtDefaultValue);
        }

        #endregion

        #region Culture dependent conversion from string to type with default value

        public static bool DefaultToIfNullOrEmpty(string strValue, bool bDefaultValue)
        {
            if (String.IsNullOrEmpty(strValue)) return bDefaultValue;
            bool bRet = bDefaultValue;
            try
            {
                bRet = Localization.ToBoolean(strValue);
            }
            catch
            {
            }
            return bRet;
        }
        public static int DefaultToIfNullOrEmpty(string strValue, int intDefaultValue)
        {
            if (String.IsNullOrEmpty(strValue)) return intDefaultValue;
            int intRet = intDefaultValue;
            try
            {
                intRet = Localization.ToInt32(strValue);
            }
            catch
            {
            }
            return intRet;
        }
        public static long DefaultToIfNullOrEmpty(string strValue, long longDefaultValue)
        {
            if (String.IsNullOrEmpty(strValue)) return longDefaultValue;
            long longRet = longDefaultValue;
            try
            {
                longRet = Localization.ToInt64(strValue);
            }
            catch
            {
            }
            return longRet;
        }
        public static double DefaultToIfNullOrEmpty(string strValue, double dblDefaultValue)
        {
            if (String.IsNullOrEmpty(strValue)) return dblDefaultValue;
            double dblRet = dblDefaultValue;
            try
            {
                dblRet = Localization.ToDouble(strValue);
            }
            catch
            {
            }
            return dblRet;
        }
        public static string DefaultToIfNullOrEmpty(string strValue, string strDefaultValue)
        {
            if (String.IsNullOrEmpty(strValue)) return strDefaultValue;

            strValue = strValue.Trim();
            if (strValue.Length <= 0) return strDefaultValue;

            return strValue;
        }
        public static DateTime DefaultToIfNullOrEmpty(string strValue, DateTime dtDefaultValue)
        {
            if (String.IsNullOrEmpty(strValue)) return dtDefaultValue;
            DateTime dtRet = dtDefaultValue;
            try
            {
                dtRet = Localization.ToDateTime(strValue);
            }
            catch
            {
            }
            return dtRet;
        }

        #endregion

        #region Miscellaneous

        public static string GetLocalizedFileName(
            string strDefaultFileName,
            string strUICultureName)
        {
            string strExtension = Path.GetExtension(strDefaultFileName);
            string strNameWithoutExtension = Path.Combine(
                Path.GetDirectoryName(strDefaultFileName),
                Path.GetFileNameWithoutExtension(strDefaultFileName));

            string strFileName = String.Format(
                "{0}.{1}{2}",
                strNameWithoutExtension,
                strUICultureName,
                strExtension);

            return strFileName;
        }

        public static string GetLocalizedWebFileName(
            string strDefaultFileName,
            string strUICultureName)
        {
            string strFileName = GetLocalizedFileName(
                strDefaultFileName,
                strUICultureName);

            strFileName = strFileName.Replace(Path.DirectorySeparatorChar, '/');

            return strFileName;
        }

        #endregion
    }
}
