// Bitground Software common helpers library
// http://www.bitground.com/
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Data;

namespace Bitground.Common.Helpers
{
    public abstract class Tables
    {
        #region Column additions

        public static void AddSelectedColumn(DataTable dt, string strIdColumn, int intSelectedId)
        {
            const string strSelected = "Selected";
            dt.Columns.Add(strSelected, typeof(bool));
            foreach (DataRow dr in dt.Rows)
            {
                int intId = Localization.ConvertToInvariantInt32(dr[strIdColumn]);
                if (intId == intSelectedId)
                {
                    dr[strSelected] = true;
                }
                else
                {
                    dr[strSelected] = false;
                }
            }
        }

        public static void AddSelectedColumn(DataTable dt, int intSelectedId)
        {
            AddSelectedColumn(dt, "Id", intSelectedId);
        }

        public static void AddSelectedColumn(DataTable dt, string strIdColumn, string strSelectedId)
        {
            const string strSelected = "Selected";
            dt.Columns.Add(strSelected, typeof(bool));
            foreach (DataRow dr in dt.Rows)
            {
                string strId = Localization.ConvertToInvariantString(dr[strIdColumn]);
                if (strId == strSelectedId)
                {
                    dr[strSelected] = true;
                }
                else
                {
                    dr[strSelected] = false;
                }
            }
        }

        public static void AddSelectedColumn(DataTable dt, string strSelectedId)
        {
            AddSelectedColumn(dt, "Id", strSelectedId);
        }

        #endregion

        #region Table manipulation

        public static DataTable Rotate(DataTable dt, string strColumNamesRow)
        {
            if (null == dt)
            {
                return dt;
            }

            DataTable dtRotated = new DataTable(dt.TableName + "Rotated");

            string strRotatedColumn = strColumNamesRow + "Rotated";
            dtRotated.Columns.Add(strRotatedColumn);
            foreach (DataRow dr in dt.Rows)
            {
                dtRotated.Columns.Add(dr[strColumNamesRow].ToString(), typeof(string));
            }

            foreach (DataColumn dc in dt.Columns)
            {
                if (dc.ColumnName != strColumNamesRow)
                {
                    DataRow drNew = dtRotated.NewRow();
                    drNew[strRotatedColumn] = dc.ColumnName;

                    foreach (DataRow dr in dt.Rows)
                    {
                        drNew[dr[strColumNamesRow].ToString()] = dr[dc.ColumnName];
                    }

                    dtRotated.Rows.Add(drNew);
                }
            }

            return dtRotated;
        }

        public static DataTable Reverse(DataTable dt)
        {
            if (null == dt)
            {
                return dt;
            }

            DataTable dtReverse = dt.Clone();

            int intRowsCount = dt.Rows.Count;
            for (int i = 0; i < intRowsCount; i++)
            {
                DataRow dr = dtReverse.NewRow();

                foreach (DataColumn dc in dt.Columns)
                {
                    dr[dc.ColumnName] = dt.Rows[intRowsCount - 1 - i][dc.ColumnName];
                }

                dtReverse.Rows.Add(dr);
            }

            return dtReverse;
        }

        public static DataTable Project(DataTable dt, params string[] arColumns)
        {
            if (null == dt)
            {
                // the projection of a null table is also a null table
                return null;
            }

            if (null == arColumns)
            {
                // we do not have any projection columns, so we'll return null
                return null;
            }

            DataTable dtRet = dt.Copy();

            foreach (DataColumn dc in dt.Columns)
            {
                string strColumnName = dc.ColumnName;

                if (0 <= Array.IndexOf(arColumns, strColumnName))
                {
                    // the column should be in the projected table, so we'll do nothing
                }
                else
                {
                    // the column is not in the projected table
                    // so we'll delete it from the returned table
                    dtRet.Columns.Remove(strColumnName);
                }
            }

            for (int i = 0; i < arColumns.Length; i++)
            {
                string c = arColumns[i];
                if (dtRet.Columns.Contains(c))
                {
                    dtRet.Columns[c].SetOrdinal(i);
                }
                else
                {
                    throw new CException(
                        "Column '{0}' was not found",
                        c);
                }
            }

            return dtRet;
        }

        #endregion

        #region Sorting

        public static DataTable Sort(DataTable dt, string strSortCriteria)
        {
            DataTable dtSorted = null;

            if (null != dt)
            {
                DataTable dtSortedTemp = dt.Copy();
                dtSortedTemp.DefaultView.Sort = strSortCriteria;

                {
                    dtSorted = dt.Clone();

                    int intRowsCount = dtSortedTemp.Rows.Count;
                    for (int i = 0; i < intRowsCount; i++)
                    {
                        DataRow dr = dtSorted.NewRow();

                        foreach (DataColumn dc in dtSortedTemp.Columns)
                        {
                            dr[dc.ColumnName] = dtSortedTemp.DefaultView[i][dc.ColumnName];
                        }

                        dtSorted.Rows.Add(dr);
                    }
                }
            }
            
            return dtSorted;
        }

        public static DataTable SortAscending(DataTable dt, string strSortColumn)
        {
            return Sort(
                dt, 
                String.Format(
                    "{0} ASC",
                    strSortColumn));
        }

        public static DataTable SortDescending(DataTable dt, string strSortColumn)
        {
            return Sort(
                dt,
                String.Format(
                    "{0} DESC",
                    strSortColumn));
        }

        #endregion

        #region Conversions and export

        public static string StringFromDataTable(DataTable dt)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("DataTable ");
            if (null != dt)
            {
                if (String.IsNullOrEmpty(dt.TableName))
                {
                    sb.Append("<untitled>");
                }
                else
                {
                    sb.Append("\"" + dt.TableName + "\"");
                }
                sb.AppendLine();

                int intColumnsNumber = dt.Columns.Count;
                int[] arMaxWidths = new int[intColumnsNumber];
                for (int i = 0; i < intColumnsNumber; i++)
                {
                    string strColumnName = dt.Columns[i].ColumnName;
                    sb.Append(strColumnName);

                    int intColumnMaxLength = MaxDBColumnLength(dt, strColumnName);
                    sb.Append(new String(' ', intColumnMaxLength - strColumnName.Length));
                    arMaxWidths[i] = intColumnMaxLength;
                }
                sb.AppendLine();

                for (int i = 0; i < intColumnsNumber; i++)
                {
                    string strColumnName = dt.Columns[i].ColumnName;
                    sb.Append(new String('-', strColumnName.Length));
                    sb.Append(new String(' ', arMaxWidths[i] - strColumnName.Length));
                }
                sb.AppendLine();

                int intRowsNumber = dt.Rows.Count;
                for (int i = 0; i < intRowsNumber; i++)
                {
                    DataRow dr = dt.Rows[i];
                    for (int k = 0; k < intColumnsNumber; k++)
                    {
                        string strValue = GetDBObjectValue(dr[k]);
                        sb.Append(strValue);
                        sb.Append(new String(' ', arMaxWidths[k] - strValue.Length));
                    }
                    sb.AppendLine();
                }
            }
            else
            {
                sb.Append("<null>");
            }
            return sb.ToString();
        }

        public static string StringFromDataTableShort(DataTable dt)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("DataTable ");
            if (null != dt)
            {
                if (String.IsNullOrEmpty(dt.TableName))
                {
                    sb.Append("<untitled>");
                }
                else
                {
                    sb.Append("\"" + dt.TableName + "\"");
                }
                sb.Append(" => ");

                int intColumnsNumber = dt.Columns.Count;
                int intRowsNumber = dt.Rows.Count;
                sb.AppendFormat(
                    "[SHORT: {0} row{1} and {2} column{3}]",
                    intRowsNumber,
                    (1 == intRowsNumber) ? "" : "s",
                    intColumnsNumber,
                    (1 == intColumnsNumber) ? "" : "s");
                sb.AppendLine();
            }
            else
            {
                sb.Append("<null>");
            }
            return sb.ToString();
        }

        const string m_strHtmlNullString = "";
        private static string GetHtmlObjectValue(object ob)
        {
            if (System.DBNull.Value == ob)
            {
                return m_strHtmlNullString;
            }
            return Localization.ToString(ob);
        }

        public static string HtmlFromDataTable(DataTable dt)
        {
            StringBuilder sb = new StringBuilder();
            if (null != dt)
            {
                sb.AppendLine("<table>");

                int intColumnsNumber = dt.Columns.Count;

                if (0 < intColumnsNumber)
                {
                    sb.Append("<thead><tr>");
                    for (int i = 0; i < intColumnsNumber; i++)
                    {
                        string strColumnName = dt.Columns[i].ColumnName;

                        sb.AppendFormat(
                            "<th>{0}</th>",
                            Html.Encode(strColumnName));
                    }
                    sb.AppendLine("</tr></thead>");
                }

                int intRowsNumber = dt.Rows.Count;
                if (0 < intRowsNumber)
                {
                    sb.AppendLine("<tbody>");
                    for (int i = 0; i < intRowsNumber; i++)
                    {
                        DataRow dr = dt.Rows[i];

                        sb.AppendFormat(
                            "<tr class=\"{0}\">",
                            (0 == (i % 2)) ? "even" : "odd");
                        for (int k = 0; k < intColumnsNumber; k++)
                        {
                            string strValue = GetHtmlObjectValue(dr[k]);
                            sb.AppendFormat(
                                "<td class=\"{0}\">{1}</td>",
                                ((0 == (k % 2)) ? "even" : "odd"),
                                Html.Encode(strValue));
                        }
                        sb.AppendLine("</tr>");
                    }
                    sb.AppendLine("</tbody>");
                }

                sb.AppendLine("</table>");
            }
            return sb.ToString();
        }

        private static string GetCVSFromCell(string strValue)
        {
            bool bQuote = false;

            // surround with quotes if it contains comma, quote or new line
            if ((strValue.Contains(",")) ||
                (strValue.Contains("\"")) ||
                (strValue.Contains("\r")) ||
                (strValue.Contains("\n")))
            {
                bQuote = true;
            }

            // double the quotes
            strValue = strValue.Replace("\"", "\"\"");

            // make new lines consistent: \r\n
            strValue.Replace("\r\n", "\r").Replace("\n", "\r").Replace("\r", "\r\n");

            if (bQuote)
            {
                strValue = String.Format(
                    "\"{0}\"",
                    strValue);
            }

            return strValue;
        }

        const string m_strCSVNullString = "";
        private static string GetCSVObjectValue(object ob)
        {
            if (System.DBNull.Value == ob)
            {
                return m_strCSVNullString;
            }
            return Localization.ToString(ob);
        }

        public static string CSVFromDataTable(DataTable dt)
        {
            if (null != dt)
            {
                int intRowsCount = dt.Rows.Count;
                int intColumnsCount = dt.Columns.Count;
                if ((0 < intRowsCount) && (0 < intColumnsCount))
                {
                    StringBuilder sb = new StringBuilder();

                    for (int i = 0; i < intRowsCount; i++)
                    {
                        DataRow dr = dt.Rows[i];

                        for (int k = 0; k < intColumnsCount; k++)
                        {
                            string strValue = GetCVSFromCell(GetCSVObjectValue(dr[k]));

                            sb.Append(strValue);
                            if (k < intColumnsCount - 1)
                            {
                                sb.Append(",");
                            }
                        }
                        sb.AppendLine();
                    }

                    return sb.ToString();
                }
            }
            return "";
        }

        public static DataTable ToStringColumnsOnly(DataTable dt)
        {
            if (null == dt)
            {
                return null;
            }

            DataTable dtRet = new DataTable();
            foreach (DataColumn dc in dt.Columns)
            {
                dtRet.Columns.Add(
                    dc.ColumnName,
                    typeof(string));
            }

            foreach (DataRow dr in dt.Rows)
            {
                DataRow drNew = dtRet.NewRow();
                foreach (DataColumn dc in dt.Columns)
                {
                    drNew[dc.ColumnName] = Strings.GetDBInvariantValue(dr[dc]);
                }
                dtRet.Rows.Add(drNew);
            }

            return dtRet;
        }

        public static int[] ToArrayIDs(DataTable dt, string strColumnName)
        {
            ArrayList ar = new ArrayList();
            foreach (DataRow dr in dt.Rows)
            {
                ar.Add(Strings.GetDBInvariantInt(dr[strColumnName]));
            }
            return (int[])ar.ToArray(typeof(int));
        }

        #endregion

        #region Rows deletion

        public static void DeleteRowsByColumn(ref DataTable dt, string strColumnName, object obValue)
        {
#if DEBUG
            if (null == dt)
            {
                throw new CException("DataTable is null in Tables.DeleteRowsByColumn");
            }

            if (!dt.Columns.Contains(strColumnName))
            {
                throw new CException("Column '{0}' was not found in Tables.DeleteRowsByColumn", strColumnName);
            }
#endif

            DataRow[] dtRows = null;
            GetRowsByColumn(dt, strColumnName, obValue, out dtRows);
            if (null != dtRows)
            {
                foreach(DataRow dr in dtRows)
                {
                    dt.Rows.Remove(dr);
                }
            }
        }

        public static void DeleteRowsByExpression(ref DataTable dt, string strExpression)
        {
#if DEBUG
            if (null == dt)
            {
                throw new CException("DataTable is null in Tables.DeleteRowsByExpression");
            }
#endif
            DataRow[] dtRows = null;
            GetRowsByExpression(dt, strExpression, out dtRows);
            if (null != dtRows)
            {
                foreach (DataRow dr in dtRows)
                {
                    dt.Rows.Remove(dr);
                }
            }
        }

        #endregion

        #region Rows retrieval

        public static void GetRowsByColumn(DataTable dt, string strColumnName, object obValue, out DataRow[] rowsFound)
        {
#if DEBUG
            if (null == dt)
            {
                throw new CException("DataTable is null in Tables.GetRowsByColumn");
            }

            if (!dt.Columns.Contains(strColumnName))
            {
                throw new CException("Column '{0}' was not found in Tables.GetRowsByColumn", strColumnName);
            }
#endif
            string strExpression = "";
            Type columnType = dt.Columns[strColumnName].DataType;
            if (columnType.Equals(typeof(System.String)))
            {
                strExpression = string.Format("{0} = '{1}'", strColumnName, (string)obValue);
            }
            else
            {
                strExpression = string.Format("{0} = {1}", strColumnName, obValue);
            }

            GetRowsByExpression(dt, strExpression, out rowsFound);
        }

        public static void GetRowsByExpression(DataTable dt, string strExpression, out DataRow[] rowsFound)
        {
            rowsFound = dt.Select(strExpression);
        }

        public static DataRow FindRowByColumnValue(DataTable dt, string strColumnName, object obValue)
        {
            DataRow[] dtRows = null;
            GetRowsByColumn(dt, strColumnName, obValue, out dtRows);

            if (0 < dtRows.Length)
            {
                return dtRows[0];
            }
            else
            {
                return null;
            }
        }

        #endregion

        #region Miscellaneous

        public static void CopyRow(DataRow drSource, DataRow drDestination)
        {
            if (null == drSource)
            {
                throw new CException("drSource cannot be null");
            }
            if (null == drDestination)
            {
                throw new CException("drDestination cannot be null");
            }

            foreach (DataColumn dc in drSource.Table.Columns)
            {
                drDestination[dc.ColumnName] = drSource[dc.ColumnName];
            }
        }

        public static long GetSize(DataTable dt)
        {
            long size = 0L;
            if ((null == dt) || (null == dt.Columns) || (null == dt.Rows))
            {
                return size;
            }

            int columns = dt.Columns.Count;
            int rows = dt.Rows.Count;

            if ((0 == columns) || (0 == rows))
            {
                return size;
            }

            for (int k = 0; k < columns; k++)
            {
                Type t = dt.Columns[k].DataType;

                bool bProcessed = false;

                if ((!bProcessed) && (typeof(byte) == t))
                {
                    bProcessed = true;
                    size += rows * sizeof(byte);
                }

                if ((!bProcessed) && (typeof(int) == t))
                {
                    bProcessed = true;
                    size += rows * sizeof(int);
                }

                if ((!bProcessed) && (typeof(bool) == t))
                {
                    bProcessed = true;
                    size += rows * sizeof(bool);
                }

                if ((!bProcessed) && (typeof(long) == t))
                {
                    bProcessed = true;
                    size += rows * sizeof(long);
                }

                if ((!bProcessed) && (typeof(short) == t))
                {
                    bProcessed = true;
                    size += rows * sizeof(short);
                }

                if ((!bProcessed) && (typeof(DateTime) == t))
                {
                    bProcessed = true;
                    size += rows * sizeof(long);
                }

                if ((!bProcessed) && (typeof(double) == t))
                {
                    bProcessed = true;
                    size += rows * sizeof(double);
                }

                if ((!bProcessed) && (typeof(float) == t))
                {
                    bProcessed = true;
                    size += rows * sizeof(float);
                }

                if ((!bProcessed) && (typeof(string) == t))
                {
                    bProcessed = true;

                    for (int i = 0; i < rows; i++)
                    {
                        DataRow dr = dt.Rows[i];

                        string s = dr[k] as string;

                        if (String.IsNullOrEmpty(s))
                        {
                            size += sizeof(int);
                        }
                        else
                        {
                            size += s.Length + sizeof(int);
                        }
                    }
                }

                if (!bProcessed)
                {
                    bProcessed = true;

                    for (int i = 0; i < rows; i++)
                    {
                        DataRow dr = dt.Rows[i];

                        if (dr[k] is string)
                        {
                            string s = dr[k] as string;

                            if (String.IsNullOrEmpty(s))
                            {
                                size += sizeof(int);
                            }
                            else
                            {
                                size += s.Length + sizeof(int);
                            }
                        }
                        else
                        {
                            // we couldn't identify the type of the cell
                            // so we'll just assume it to be of the size of an int
                            size += sizeof(int);
                        }
                    }
                }
            }

            return size;
        }

        public static object GetRowValueByKey(DataTable dt, string strKeyColumnName, object obKeyValue, string strValueColumnName)
        {
#if DEBUG
            if (null == dt)
            {
                throw new CException("DataTable is null in Tables.GetRowValueByKey");
            }

            if (!dt.Columns.Contains(strKeyColumnName))
            {
                throw new CException("Column '{0}' was not found in Tables.GetRowValueByKey", strKeyColumnName);
            }

            if (!dt.Columns.Contains(strValueColumnName))
            {
                throw new CException("Column '{0}' was not found in Tables.GetRowValueByKey", strValueColumnName);
            }
#endif

            DataRow dr = FindRowByColumnValue(dt, strKeyColumnName, obKeyValue);

            if (null == dr)
            {
                return null;
            }

            return dr[strValueColumnName];
        }

        const string m_strDBNullString = "<null>";
        private static string GetDBObjectValue(object ob)
        {
            if (System.DBNull.Value == ob)
            {
                return m_strDBNullString;
            }
            return Localization.ToString(ob);
        }

        private static int MaxDBColumnLength(DataTable dt, string strColumnName)
        {
            int intMax = 0;
            intMax = Math.Max(intMax, strColumnName.Length);
            intMax = Math.Max(intMax, m_strDBNullString.Length);
            foreach (DataRow dr in dt.Rows)
            {
                intMax = Math.Max(intMax, GetDBObjectValue(dr[strColumnName]).Length);
            }
            return intMax + 2;
        }

        public static int CountDistinct(DataTable dt, string strColumnName)
        {
            DataView dataView = new DataView(dt);
            DataTable dtNew = dataView.ToTable(true, strColumnName);
            return dtNew.Rows.Count;
        }

        #endregion
    }
}
