﻿// zkBox C# client
// http://www.zkbox.com/developers/client/csharp
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using Rest;
using Rest.Service;
using Rest.Transport;
using Bitground.Common.Helpers;
using Debug = Bitground.Common.Helpers.Debug;

namespace Bitground.zkBox.Client.CS
{
    public partial class Handler
    {
        internal delegate U AuthenticateApplicationF<U>(Dictionary<string, object> h);
        internal delegate U AuthenticateApplicationAndUserF<U>(Dictionary<string, object> h);

        internal static string LogClientContext(string operation)
        {
            return string.Format(
                "ZKBOX CLIENT CS: {0}",
                operation);
        }

        internal U Authenticate_Application<U, D>(
                string method,
                string resource,
                object[] args,
                AuthenticateApplicationF<U> fn)
            where U : WSRet<D>, new()
            where D : new()
        {
            return Authenticate_AuthenticateCore<U, D>(
                method, 
                resource,
                args,
                fn, 
                null);
        }

        internal U Authenticate_ApplicationAndUser<U, D>(
                string method,
                string resource,
                object[] args,
                AuthenticateApplicationAndUserF<U> fn)
            where U : WSRet<D>, new()
            where D : new()
        {
            return Authenticate_AuthenticateCore<U, D>(
                method, 
                resource,
                args,
                null, 
                fn);
        }

        private static bool LogInput(params object[] args)
        {
            if (ParamsLogging.CanLogParams)
            {
                return ParamsLogging.LogParams("IN", args);
            }
            else
            {
                return true;
            }
        }

        private static T LogOutput<T>(Stopwatch stopwatch, T ret)
        {
            stopwatch.Stop();

            if (ParamsLogging.CanLogParams)
            {
                TimeSpan elapsedTime = stopwatch.Elapsed;

                ParamsLogging.LogParams(
                    "OUT",

                    String.Format(
                        "[TIME TAKEN {0}]",
                        Time.TimeSpanAsStringAccurateHigh(elapsedTime)),
                    "",

                    "",
                    ret);
            }

            return ret;
        }

        private U Authenticate_AuthenticateCore<U, D>(
                string method,
                string resource,
                object[] args,
                AuthenticateApplicationF<U> fnAuthenticateApplication,
                AuthenticateApplicationAndUserF<U> fnAuthenticateApplicationAndUser)
            where U : WSRet<D>, new()
            where D : new()
        {
            U ret = new U()
            {
                error = 0,   // Success
                data = default(D)
            };

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            Debug.Assert(
                ((null == fnAuthenticateApplication) && (null != fnAuthenticateApplicationAndUser))
                ||
                ((null != fnAuthenticateApplication) && (null == fnAuthenticateApplicationAndUser))
            );

            if (0 != args.Length % 2)
            {
                Log.Error("We should have an even number of arguments");
                throw new ArgumentException("We should have an even number of arguments");
            }

            RestParameterCollection parameters = new RestParameterCollection();

            for (int i = 0; i < args.Length; i += 2)
            {
                string name = args[i] as string;
                string value = (null == args[i + 1]) ? "<null>" : args[i + 1] as string;

                if ((null == name) || (null == value))
                {
                    if (null == name)
                    {
                        Log.Error("Parameter #{0} should be a string", i);
                        throw new ArgumentException(String.Format("We should have an even number of arguments", i));
                    }
                    if (null == value)
                    {
                        Log.Error("Parameter #{0} should be a string", i + 1);
                        throw new ArgumentException(String.Format("We should have an even number of arguments", i + 1));
                    }
                }
                else
                {
                    parameters.Add(name, value);
                }
            }

            RestTransport transport;
            switch (method)
            {
                case "GET":
                    transport = new GetRestTransport();
                    break;

                case "POST":
                    transport = new PostRestTransport();
                    break;

                case "PUT":
                    transport = new PutRestTransport();
                    break;

                case "DELETE":
                    transport = new DeleteRestTransport();
                    break;

                default:
                    Log.Error("Invalid method");
                    throw new ArgumentException("Invalid method");
            }

            using (Log.PushContext(method + " " + resource))
            {
                if (!LogInput(args))
                {
                    Log.Error("Failure while logging the input parameters");
                    ret.error = 1;     // System error
                }
                else
                {
                    bool userAuthenticate = false;

                    string userId = null;
                    string userVersion = null;
                    string userSessionId = null;
                    string userSessionKey = null;

                    if (null != fnAuthenticateApplicationAndUser)
                    {
                        if (!status.isAuthenticated)
                        {
                            ret.error = 2; // Not authenticated
                        }
                        else
                        {
                            userAuthenticate = true;
                            userId = status.userId;
                            userVersion = status.userVersion;
                            userSessionId = status.sessionId;
                            userSessionKey = status.sessionKey;
                        }
                    }


                    if (0 == ret.error)
                    {
                        Hashtable h = null;
                        string zkBoxReference = null;
                        string zkBoxRuntime = null;
                        try
                        {
                            h = proxy.Invoke(
                                transport,
                                resource,
                                userAuthenticate,
                                userId,
                                userVersion,
                                userSessionId,
                                userSessionKey,
                                parameters,
                                out zkBoxReference,
                                out zkBoxRuntime);
                        }
                        catch (Exception ex)
                        {
                            Log.Exception(ex, "Failure while calling the method");
                        }

                        if (null != zkBoxReference)
                        {
                            ret.reference = zkBoxReference;
                            if (Log.CanLogTrace) Log.Trace("We have the response header X-zkBox-Reference: {0}", zkBoxReference);
                        }
                        if (null != zkBoxRuntime)
                        {
                            ret.runtime = zkBoxRuntime;
                            if (Log.CanLogTrace) Log.Trace("We have the response header X-zkBox-Runtime: {0}", zkBoxRuntime);
                        }

                        if (null != h)
                        {
                            int error = Nulls.Int32;
                            if (null != h["error"])
                            {
                                try
                                {
                                    error = (int)Math.Round(Double.Parse(h["error"].ToString()));
                                }
                                catch
                                {
                                    // do nothing
                                }
                            }

                            if (0 == error)
                            {
                                var data = h["data"] as Dictionary<string, object>;

                                if (null != data)
                                {
                                    string reference = ret.reference;
                                    string runtime = ret.runtime;
                                    if (null != fnAuthenticateApplication)
                                    {
                                        ret = fnAuthenticateApplication(data);
                                    }
                                    else
                                    {
                                        ret = fnAuthenticateApplicationAndUser(data);
                                    }
                                    ret.reference = reference;
                                    ret.runtime = runtime;
                                }
                                else
                                {
                                    Log.Error("The api call returned a semi-valid response. Although the error returned is saying success, there was no data returned");
                                    ret.error = 1;     // System error
                                }
                            }
                            else
                            {
                                if (Nulls.Int32 == error)
                                {
                                    Log.Error("The api call returned an invalid response. The error field is missing or is invalid");
                                    ret.error = 1;     // System error
                                }
                                else
                                {
                                    string message = "";
                                    if (null != h["message"])
                                    {
                                        message = h["message"].ToString();
                                    }

                                    if (String.IsNullOrEmpty(message))
                                    {
                                        Log.Error("A logical error occured when calling the method. The error was: '{0}'. OTOH the message field is missing.", error);
                                        ret.error = error;
                                    }
                                    else
                                    {
                                        if (Log.CanLogInfo) Log.Info("An logical error occured when calling the method. The error is: '{0}'. The error message is: '{1}'", error, message);
                                        ret.error = error;
                                        ret.message = message;
                                    }
                                }
                            }
                        }
                        else
                        {
                            Log.Error("The api call returned a null value");
                            ret.error = 1;     // System error
                        }
                    }
                }
            }

            return ret;
        }
    }
}
