﻿// zkBox C# client
// http://www.zkbox.com/developers/client/csharp
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using Bitground.Common.Helpers;
using Debug = Bitground.Common.Helpers.Debug;

namespace Bitground.zkBox.Client.CS
{
    public partial class Handler : IDisposable
    {
        private Proxy proxy;
        public Applications Application;
        public Users Users;
        public Objects Objects;
        public Transactions Transactions;
        public Batch Batch;
        private Handler()
        {
        }

        internal Status status;

        private string m_apiUrl;
        private string m_apiVersion;
        private string m_applicationId;
        private string m_applicationKey;

        internal bool hasCache;
        internal Cache cache = null;

        public class Init
        {
            public string apiUrl = null;
            public string apiVersion = null;
            public string applicationId = null;
            public string applicationKey = null;

            public bool hasCache = true;

            internal void CheckParameters()
            {
                Debug.Assert(!String.IsNullOrEmpty(apiUrl), "The initialization parameter apiUrl is null or it wasn't provided");
                Debug.Assert(!String.IsNullOrEmpty(apiVersion), "The initialization parameter apiVersion is null or it wasn't provided");
                Debug.Assert(!String.IsNullOrEmpty(applicationId), "The initialization parameter applicationId is null or it wasn't provided");
                Debug.Assert(!String.IsNullOrEmpty(applicationKey), "The initialization parameter applicationKey is null or it wasn't provided");
            }

            public override string ToString()
            {
                return String.Format(
                    "[apiUrl='{0}' apiVersion='{1}' applicationId='{2}' applicationKey='{3}' hasCache='{4}']",
                    apiUrl,
                    apiVersion,
                    applicationId,
                    applicationKey,
                    hasCache);
            }
        }

        public Handler(
            Init init)
        {
            Application = new Applications(this);
            Users = new Users(this);
            Objects = new Objects(this);
            Transactions = new Transactions(this);
            Batch = new Batch(this);

            string referenceId;
            using (Log.PushRootContext("Global", out referenceId))
            {
                TrueRandom.Initialize();
            }

            using (Log.PushRootContext(LogClientContext("START"), out referenceId))
            {
                init.CheckParameters();

                m_apiUrl = init.apiUrl;
                m_apiVersion = init.apiVersion;
                m_applicationId = init.applicationId;
                m_applicationKey = init.applicationKey;
                hasCache = init.hasCache;
                if (Log.CanLogInfo) Log.Info(
                    "Initializing with the following parameters: {0}",
                    init.ToString());

                proxy = new Proxy(init);
                status = new Status();
                if (hasCache)
                {
                    cache = new Cache();
                }
            }
        }

        public void Dispose()
        {
            string referenceId;
            using (Log.PushRootContext(LogClientContext("STOP"), out referenceId))
            {
                TrueRandom.Dispose();
                if (Log.CanLogInfo) Log.Info("The api client was stopped");
            }
        }


        internal delegate U CallF<U>();

        internal U Call<U, D>(
                string method,
                object[] args,
                CallF<U> fnCall)
            where U : WSRet<D>, new()
            where D : new()
        {
            string referenceId;
            using (Log.PushRootContext(LogClientContext("CALL"), out referenceId))
            {
                try
                {
                    LogInput(new string[] { method + ":", "" }.Concat(args).ToArray());

                    Stopwatch stopwatch = new Stopwatch();
                    stopwatch.Start();


                    return LogOutput(stopwatch, fnCall());
                }
                catch (Exception ex)
                {
                    Log.Exception(ex, "Failure during the Call");
                    return null;
                }
            }
        }
    }
}
