// zkBox C# client
// http://www.zkbox.com/developers/client/csharp
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;
using System.Xml.XPath;
using System.Net;
using System.Security.Cryptography;
using Rest;
using Rest.Xml; 
using Rest.Service;
using Rest.Transport;
using Bitground.Common.Helpers;

namespace Bitground.zkBox.Client.CS
{
    /// <summary>
    /// Base class for zkBox web service proxy
    /// </summary>
    /// <remarks>
    /// Authenticating REST requests
    /// </remarks>
    internal class ProxyBase : RestServiceProxyBase, IRestServiceXmlSchema
    {
        #region Properties

        private string m_apiUrl;
        private string m_apiVersion;
        private string m_applicationId;
        private string m_applicationKey;

        private int m_maxErrorRetry;

        private string m_apiHost;

        #endregion

        private ProxyBase() { }

        public ProxyBase(
            Handler.Init init)
        {
            m_apiUrl = init.apiUrl;

            {
                var uri = new Uri(m_apiUrl);
                m_apiHost = uri.Host;
            }

            switch (init.apiVersion)
            {
                case "1.0":                 // api version 1.0
                    m_apiVersion = "0100";
                    break;

                default:
                    throw new ArgumentException("Invalid apiVersion");
            }

            m_applicationId = init.applicationId;
            m_applicationKey = init.applicationKey;

            m_maxErrorRetry = 8;    // maximum retry backoff delay will be 2^8*17 ~ 4.3sec
        }

        private void PauseOnRetry(int retries, HttpStatusCode status)
        {
            if (retries <= m_maxErrorRetry)
            {
                int maxdelay = (int)Math.Pow(2, retries) * 17;
                int delay = 1 + new Random((int)DateTime.Now.Ticks).Next() % maxdelay;
                if (Log.CanLogTrace) Log.Trace("PauseOnRetry (retries={0}, status={1} (code {2}): {3}ms", 
                    retries,
                    status,
                    (int)status,
                    delay);
                System.Threading.Thread.Sleep(delay);
            }
            else
            {
                throw new CException("Maximum number of retry attempts reached : {0}; status is {1} (code {2}))", (retries - 1), status, (int)status);
            }
        }

        private string InvokeCore(RestCommand restCommand, out WebHeaderCollection responseHeaders)
        {
            // when the operation fails, retry it, if possible

            responseHeaders = null;
            string response = null;
            bool shouldRetry = true;
            int retries = 0;
            do
            {
                try
                {
                    response = base.Invoke(restCommand, out responseHeaders);
                    shouldRetry = false;
                }
                catch (WebException wex)
                {
                    shouldRetry = false;

                    using (HttpWebResponse httpErrorResponse = (HttpWebResponse)wex.Response as HttpWebResponse)
                    {
                        if (httpErrorResponse == null)
                        {
                            Log.Exception(wex, "WebException in InvokeCore (null response)");
                            throw;
                        }

                        if ((HttpStatusCode.InternalServerError == httpErrorResponse.StatusCode) ||
                            (HttpStatusCode.ServiceUnavailable == httpErrorResponse.StatusCode))
                        {
                            shouldRetry = true;
                            PauseOnRetry(++retries, httpErrorResponse.StatusCode);
                        }
                        else
                        {
                            Log.Exception(wex, "WebException in InvokeCore");
                            throw;
                        }
                    }
                }
                catch (Exception ex)
                {
                    shouldRetry = false;
                    Log.Exception(ex, "Exception in InvokeCore");
                    throw;
                }
            } while (shouldRetry);

            return response;
        }

        internal Hashtable Invoke(
            RestTransport transport, 
            string resource,
            bool userAuthenticate,
            string userId,
            string userVersion,
            string userSessionId,
            string userSessionKey,
            RestParameterCollection parameters,
            out string zkBoxReference,
            out string zkBoxRuntime)
        {
            resource = "/" + resource;

            RestCommand command = new RestCommand(m_apiUrl + resource, transport);

            string signature = CreateSignature(
                transport.HttpMethod,
                resource,
                command.Transport.HttpContentType.ContentType);

            if (Log.CanLogTrace) Log.Trace("Setting X-zkBox-Auth header to: {0}", signature);
            command.Headers.Add("X-zkBox-Auth", signature);

            // !!! the user agent should begin with "zkBox-Client"
            //     based on this, the zkBox is checking for cookie or for the header
            command.UserAgent = "zkBox-Client/1.0; (.Net)";      // api version 1.0

            if (true == userAuthenticate)
            {
                string userSignature = CreateUserSignature(
                    transport.HttpMethod,
                    resource,
                    Rest.Http.JsonUrlEncoded.JsonUrlEncoded.ContentType,
                    Encoding.UTF8,
                    parameters,
                    userId,
                    userVersion,
                    userSessionId,
                    userSessionKey);

                if (Log.CanLogTrace) Log.Trace("Setting X-zkBox-UserAuth header to: {0}", userSignature);
                command.Headers.Add("X-zkBox-UserAuth", userSignature);
            }

            // Add params
            if (parameters != null)
            {
                command.Parameters.Add(parameters);
            }

            WebHeaderCollection responseHeaders;
            string json = InvokeCore(command, out responseHeaders);

            zkBoxReference = null;
            zkBoxRuntime = null;
            if (null != responseHeaders)
            {
                zkBoxReference = responseHeaders["X-zkBox-Reference"] as string;
                zkBoxRuntime = responseHeaders["X-zkBox-Runtime"] as string;
            }

            return Json.Decode<Hashtable>(json);
        }

        private string ByteArrayToHexString(byte[] ar)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < ar.Length; i++)
            {
                sb.Append(ar[i].ToString("X2"));
            }
            return sb.ToString();
        }

        protected virtual string CreateSignature(
            string method,
            string resource,
            string contentType)
        {
            string contentMD5 = "";
            string now = DateTime.UtcNow.ToString("yyyyMMddHHmm");

            return CreateSignature(
                method,
                resource,
                now,
                contentType,
                contentMD5);
        }

        private string CreateSignature(
            string method,
            string resource,
            string now,
            string contentType,
            string contentMD5)
        {
            string stringToSign =
                method + "\n" +
                resource + "\n" +
                now + "\n" +
                contentType + "\n" +
                contentMD5 + "\n";

            string signature =
                Crypto.HMACHash(
                    m_applicationKey,
                    stringToSign);

            return m_apiVersion + m_applicationId + now + signature;
        }

        protected virtual string CreateUserSignature(
            string method,
            string resource,
            string contentType,
            Encoding encoding,
            RestParameterCollection parameters,
            string userId,
            string userVersion,
            string userSessionId,
            string userSessionKey)
        {
            string nonce = UniqueId.GetPseudo().ToUpper();

            string strdata = null;
            {
                MemoryStream ms = new MemoryStream();

                Rest.Http.JsonUrlEncoded.JsonUrlEncoded.WriteRequestBody(
                    ms,
                    parameters,
                    null,           // dataItems
                    encoding);

                ms.Seek(0, SeekOrigin.Begin);
                strdata = new StreamReader(ms).ReadToEnd();
            }

            string dataToSign =
                method + "\n" +
                resource + "\n" +
                nonce + "\n" +
                contentType + "\n" +
                ((null == strdata) ? "" : strdata) + "\n";

            if (Log.CanLogTrace) Log.Trace("User dataToSign is '{0}'", dataToSign);
            string userDataSignature = Crypto.HMACHash(userSessionKey, dataToSign);

            string userAuth = nonce + userDataSignature;
            // as in the zkBox.status.js / updateUserAuth method
            string userSignature = userId + userVersion + userSessionId + userAuth;

            return userSignature;
        }

        protected virtual Boolean IsErrorResponse(XmlDocument xmlDocument)
        {
            return xmlDocument.SelectSingleNode("//Response/Errors") != null;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="xmlDocument"></param>
        /// <returns></returns>
        protected virtual string GetExceptionMessage(XmlDocument xmlDocument)
        {
            return
                   "Error code: " +
                   xmlDocument.SelectSingleNode("//Response/Errors/Error/Code/text()").Value + "." +
                   xmlDocument.SelectSingleNode("//Response/Errors/Error/Message/text()").Value;
        }

        #region IRestServiceXmlSchema Members
        /// <summary>
        /// 
        /// </summary>
        /// <param name="xmlDocument"></param>
        /// <returns></returns>
        Boolean IRestServiceXmlSchema.IsErrorResponse(XmlDocument xmlDocument)
        {
            return IsErrorResponse(xmlDocument);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="xmlDocument"></param>
        /// <returns></returns>
        string IRestServiceXmlSchema.GetExceptionMessage(XmlDocument xmlDocument)
        {
            return GetExceptionMessage(xmlDocument);
        }
        #endregion
    }
}
