﻿// zkBox C# client
// http://www.zkbox.com/developers/client/csharp
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Bitground.Common.Helpers;

namespace Bitground.zkBox.Client.CS
{
    public class Objects : ApiHandler
    {
        public Objects(Handler handler)
            : base(handler)
        {
        }

        internal class ObjectItem
        {
            public string objectId;
            public string type;
            public object data;
        }

        #region Add object

        public class AddObject : WSRet<AddObject.Data>
        {
            public class Data
            {
                public string objectId;
            }
        };

        public AddObject Add(string type, object obj)
        {
            AddObject ret = api.Call<AddObject, AddObject.Data>(
                "Object.Add",
                new object[] 
                {   
                    "type", type,
                    "obj", obj
                },
                () =>
                {
                    if (Log.CanLogTrace) Log.Trace("Adding the object");

                    string objEncryptedData;
                    Crypto.SRPClient.AddObject(
                        api.status.cryptoKey,
                        Json.Encode(obj),
                        out objEncryptedData);

                    return api.Authenticate_ApplicationAndUser<AddObject, AddObject.Data>(
                        "POST",
                        (null == type) ? "object" : ("object/type/" + type),
                        new object[] 
                        { 
                            "type", type,
                            "data", objEncryptedData
                        },

                        (data) =>
                        {
                            string objectId = data["objectId"] as string;

                            if (Log.CanLogTrace) Log.Trace("The object was added");
                            return new AddObject
                            {
                                error = 0,
                                data = new AddObject.Data
                                {
                                    objectId = objectId
                                }
                            };
                        });
                });

            if ((null != ret) && (0 == ret.error) && (api.hasCache))
            {
                // if the operation succeeded, then add the object to the cache

                string objectId = ret.data.objectId;

                var ob = new ObjectItem
                {
                    objectId = ret.data.objectId,
                    type = type,
                    data = obj
                };
                api.cache.SetItem("object-" + objectId, ob);

                // all objects by type
                {
                    var objectIdsOfType = api.cache.GetItem("objects-type-" + type) as ArrayList;
                    if (null != objectIdsOfType)
                    {
                        objectIdsOfType.Add(objectId);
                        api.cache.SetItem("objects-type-" + type, objectIdsOfType);
                    }
                }

                // all objects
                {
                    var objectIds = api.cache.GetItem("objects-all") as ArrayList;
                    if (null != objectIds)
                    {
                        objectIds.Add(objectId);
                        api.cache.SetItem("objects-all", objectIds);
                    }
                }
            }

            return ret;
        }

        #endregion

        #region Get object

        public class GetObject : WSRet<GetObject.Data>
        {
            public class Data
            {
                public string objectId;
                public string type;
                public object data;
            }
        };

        public GetObject Get(string objectId)
        {
            if (api.hasCache)
            {
                var ob = api.cache.GetItem("object-" + objectId) as ObjectItem;
                if (null != ob)
                {
                    if (Log.CanLogTrace) Log.Trace("Cache hit: GetObject('{0}') will return '{1}'", objectId, Params.GetObjectAsString(ob));
                    return new GetObject
                    {
                        error = 0,
                        data = new GetObject.Data
                        {
                            objectId = ob.objectId,
                            type = ob.type,
                            data = ob.data
                        }
                    };
                }
            }

            GetObject ret = api.Call<GetObject, GetObject.Data>(
                "Objects.Get",
                new object[] 
                {   
                    "objectId", objectId
                },
                () =>
                {
                    if (Log.CanLogTrace) Log.Trace("Getting the object");

                    return api.Authenticate_ApplicationAndUser<GetObject, GetObject.Data>(
                        "GET",
                        "object/" + objectId,
                        new object[] 
                        { 
                        },
                        (data) =>
                        {
                            string objId = data["objectId"] as string;
                            string objType = data["type"] as string;
                            string objEncryptedData = data["data"] as string;

                            string objData;
                            Crypto.SRPClient.GetObject(
                                api.status.cryptoKey,
                                objEncryptedData,
                                out objData);

                            if (Log.CanLogTrace) Log.Trace("The object was retrieved");
                            return new GetObject
                            {
                                error = 0,
                                data = new GetObject.Data
                                {
                                    objectId = objId,
                                    type = objType,
                                    data = Json.Decode<object>(objData)
                                }
                            };
                        });
                });

            if ((null != ret) && (0 == ret.error) && (api.hasCache))
            {
                var ob = new ObjectItem
                {
                    objectId = ret.data.objectId,
                    type = ret.data.type,
                    data = ret.data.data
                };
                api.cache.SetItem("object-" + objectId, ob);
            }

            return ret;
        }

        #endregion

        #region Update object

        public class UpdateObject : WSRet<UpdateObject.Data>
        {
            public class Data
            {
            }
        };

        public UpdateObject Update(string objectId, object obj)
        {
            UpdateObject ret = api.Call<UpdateObject, UpdateObject.Data>(
                "Objects.Update",
                new object[] 
                {   
                    "objectId", objectId,
                    "obj", obj
                },
                () =>
                {
                    if (Log.CanLogTrace) Log.Trace("Updating the object");

                    string objEncryptedData;
                    Crypto.SRPClient.UpdateObject(
                        api.status.cryptoKey,
                        Json.Encode(obj),
                        out objEncryptedData);

                    return api.Authenticate_ApplicationAndUser<UpdateObject, UpdateObject.Data>(
                        "PUT",
                        "object/" + objectId,
                        new object[] 
                        { 
                            "data", objEncryptedData
                        },

                        (data) =>
                        {
                            if (Log.CanLogTrace) Log.Trace("The object was updated");
                            return new UpdateObject
                            {
                                error = 0,
                                data = new UpdateObject.Data
                                {
                                }
                            };
                        });
                });

            if ((null != ret) && (0 == ret.error) && (api.hasCache))
            {
                var ob = api.cache.GetItem("object-" + objectId) as ObjectItem;
                if (null != ob)
                {
                    ob.data = obj;
                    api.cache.SetItem("object-" + objectId, ob);
                }
            }

            return ret;
        }

        #endregion

        #region Delete object

        public class DeleteObject : WSRet<DeleteObject.Data>
        {
            public class Data
            {
                public string type;
            }
        };

        public DeleteObject Delete(string objectId)
        {
            DeleteObject ret = api.Call<DeleteObject, DeleteObject.Data>(
                "Objects.Delete",
                new object[] 
                {   
                    "objectId", objectId
                },
                () =>
                {
                    if (Log.CanLogTrace) Log.Trace("Deleting the object");

                    return api.Authenticate_ApplicationAndUser<DeleteObject, DeleteObject.Data>(
                        "DELETE",
                        "object/" + objectId,
                        new object[] { },

                        (data) =>
                        {
                            string type = data["type"] as string;

                            if (Log.CanLogTrace) Log.Trace("The object was deleted");
                            return new DeleteObject
                            {
                                error = 0,
                                data = new DeleteObject.Data
                                {
                                    type = type
                                }
                            };
                        });
                });

            if ((null != ret) && (0 == ret.error) && (api.hasCache))
            {
                api.cache.RemoveItem("object-" + objectId);
                string type = ret.data.type;

                // all objects by type
                {
                    var objectIdsOfType = api.cache.GetItem("objects-type-" + type) as ArrayList;
                    if (null != objectIdsOfType)
                    {
                        if (0 <= objectIdsOfType.IndexOf(objectId))
                        {
                            objectIdsOfType.Remove(objectId);
                            if (Log.CanLogInfo) Log.Info("We removed the item from the objects-type list. The new value is: {0}", Params.GetObjectAsString(objectIdsOfType));
                            api.cache.SetItem("objects-type-" + type, objectIdsOfType);
                        }
                        else
                        {
                            Log.Error("An unexpected error has occurred. We do not have the item in the objects-type list. The objectId is '{0}' and the list is: {1}", objectId, Params.GetObjectAsString(objectIdsOfType));
                        }
                    }
                }

                // all objects
                {
                    var objectIds = api.cache.GetItem("objects-all") as ArrayList;
                    if (null != objectIds)
                    {
                        if (0 <= objectIds.IndexOf(objectId))
                        {
                            objectIds.Remove(objectId);
                            if (Log.CanLogInfo) Log.Info("We removed the item from the objects-all list. The new value is: {0}", Params.GetObjectAsString(objectIds));
                            api.cache.SetItem("objects-all" + type, objectIds);
                        }
                        else
                        {
                            Log.Error("An unexpected error has occurred. We do not have the item in the objects-all list. The objectId is '{0}' and the list is: {1}", objectId, Params.GetObjectAsString(objectIds));
                        }
                    }
                }
            }

            return ret;
        }

        #endregion

        #region Get objects list

        // TODO_LATER: finish the C# client

        #endregion

        #region Get objects list IDs

        // TODO_LATER: finish the C# client

        #endregion

        #region Set single object

        // TODO_LATER: finish the C# client

        #endregion

        #region Get single object

        // TODO_LATER: finish the C# client
        
        #endregion
    }
}
