// srp4net 1.0
// SRP for .NET - A JavaScript/C# .NET library for implementing the SRP authentication protocol
// http://code.google.com/p/srp4net/
// Copyright 2010, Sorin Ostafiev (http://www.ostafiev.com/)
// License: GPL v3 (http://www.gnu.org/licenses/gpl-3.0.txt)

// Description: Secure Remote Password protocol implementation in JavaScript and C#
// RFC: http://www.ietf.org/rfc/rfc2945.txt
// Details: http://srp.stanford.edu/ndss.html
// Implementation: SRP variant 6a with SHA2 digest (instead of SHA1 from the original specification) on 512 bits
// Author: Sorin Ostafiev (http://www.ostafiev.com/)
// Libraries used:
//      - BSD: jsSHA.js - A JavaScript implementation of the SHA family of hashes, as defined in FIPS PUB 180-2
//              originally written by Brian Turek http://jssha.sourceforge.net/
//              changed by Atsushi Oka (http://ats.oka.nu/titaniumcore/js/crypto/jsSHA.js, several functions taken from Paul Johnson)
//              slightly changed by Sorin Ostafiev (http://www.ostafiev.com/)
//      - LGPL: SHA.js - SHA class that calculates various SHA hash values
//              by Atsushi Oka (http://ats.oka.nu/titaniumcore/js/crypto/SHA.js)
//              special thanks to Mr. Brian Turek (http://sourceforge.net/users/caligatio)
//      - BSD: sha512.js - A JavaScript implementation of the Secure Hash Algorithm, SHA-512, as defined in FIPS 180-2
//              Copyright Anonymous Contributor, Paul Johnston 2000 - 2009
//              Other contributors: Greg Holt, Andrew Kepert, Ydnar, Lostinet
//      - Public: BigInt.js - Big Integer Library
//              by Leemon Baird (http: //www.leemon.com/crypto/BigInt.js)
//      - LGPL: binary.js - Tools for creating, modifying binary data
//              by Atsushi Oka (http://ats.oka.nu/titaniumcore/js/tools/binary.js)
//      - MIT/GPL - jquery-1.3.2.js - jQuery - a new kind of JavaScript Library
//              by the jQuery team (http://docs.jquery.com/Contributors)

$zkBox.crypto.srp =
{
    N: '',
    g: '',
    k: '',

    Nbits: 0,

    one: '',
    two: '',
    three: '',

    bigint2hex: function(i)
    {
        // convert from a bigint number to a hex string
        return bigInt2str(i, 16);
    },

    hex2bigint: function(s)
    {
        // convert from a hex string to a bigint number
        return str2bigInt(s, 16, 1);
    },


    compareTo: function(big1, big2)
    {
        // -1 if big1 < big2
        //  1 if big1 > big2
        //  0 if equal
        if (greater(big1, big2))
        {
            return 1;
        }
        else
        {
            if (equals(big1, big2))
            {
                return 0;
            }
        }
        return -1;
    },

    to128: function(s)
    {
        // pad a string with zeroes up to length of 128
        var i;
        var l = 128 - s.length;
        for (i = 0; i < l; i++)
        {
            s = '0' + s;
        }
        return s;
    },

    H: function(x)
    {
        // hash a message
        return $zkBox.crypto.hash(x);
    },

    HHex: function(x)
    {
        // hash a hex string
        return $zkBox.crypto.hashHex(x);
    },

    CheckUser: function(username, fnProgress, fnSuccess, fnError)
    {
        // Check if a given user exists
        // username = the username entered in the user interface
        if (!this.initialized) throw "SRP not initialized";

        fnProgress('Checking for user');
        var login = this.HashUser(username);
        username = '';

        //fnProgress('Calling the Users/CheckUser server\'s method');
        fnProgress('Checking the user');
        if ($zkBox.logging) $zkBox.utils.log("login: " + login);
        // call the server's Users/CheckUser method
        $zkBox.model.Call(false, false, "PUT user/check", {
            login: login
        },
        function(data)
        {
            fnProgress('The user was checked');
            fnSuccess(data);
        },
        function(error)
        {
            fnProgress('Failure while checking the user');
            fnError(error);
        });
    },

    HashUser: function(username)
    {
        return this.H("!@#<32}|{_$+)EW:>fWS@@!=39dje%^#$RF']]ew3" + username + "shaSHA@");
    },

    HashPass: function(password)
    {
        return this.H("\":}[\|weFC@de';{{3$$  vdRF2w^5V|\\/.w32" + password + "sshHAA!%");
    },

    BeginUpdateUser: function(login, newpassword, fnProgress, fnSuccess, fnError)
    {
        if (!this.initialized) throw "SRP not initialized";

        fnProgress('SRP begin update user started');

        var newpass = this.HashPass(newpassword);
        newpassword = '';
        if ($zkBox.logging) $zkBox.utils.log('newpass:' + newpass);

        fnProgress('Salting and hashing the input');
        // s - login's salt
        // s = random number
        var s = randBigInt(this.Nbits, 0); //TODO_LATER: use TRNG
        var sHex = this.to128(this.bigint2hex(s));

        fnProgress('Compute the private key');
        // x - a private key derived from the password and salt
        // x = H(s || H(username) || H(password))
        var x;
        {
            var x1 = this.hex2bigint(this.HHex(
                sHex +
                login +
                newpass));

            if (this.compareTo(x1, this.N) < 0) { x = x1; }
            else { x = mod(x1, sub(this.N, this.one)); }
        }

        fnProgress('Generate the server\'s verifier');
        // v - the server's verifier
        // v = g^x (mod N)
        var v = powMod(this.g, x, this.N);
        var vHex = this.to128(this.bigint2hex(v));

        fnProgress('Calling the Users/BeginUpdateUser server\'s method');
        if ($zkBox.logging) $zkBox.utils.log("    s: " + sHex);
        if ($zkBox.logging) $zkBox.utils.log("    v: " + vHex);
        // call the server's Users/BeginUpdateUser method
        $zkBox.model.Call(false, true, "POST user/update", {
            s: sHex,
            v: vHex
        },
        function(data)
        {
            fnProgress('The user update was successfully started');
            fnProgress('Decoding the response');

            if ($zkBox.logging) $zkBox.utils.log("data: " + $zkBox.utils.dump(data));

            fnSuccess(
                login,
                newpass,
                data.newUserVersion);
        },
        function(error)
        {
            fnProgress('Failure during the SRP begin update user');
            fnError(error);
        });
    },

    CommitUpdateUser: function(newUserVersion, fnProgress, fnSuccess, fnError)
    {
        if (!this.initialized) throw "SRP not initialized";

        fnProgress('SRP commit update user started');

        fnProgress('Calling the Users/CommitUpdateUser server\'s method');
        if ($zkBox.logging) $zkBox.utils.log("newUserVersion: " + newUserVersion);
        // call the server's Users/CommitUpdateUser method
        $zkBox.model.Call(false, true, "PUT user/update/" + newUserVersion, {},
        function(data)
        {
            fnProgress('The user update was successfully commited');
            fnProgress('Decoding the response');

            if ($zkBox.logging) $zkBox.utils.log("data: " + $zkBox.utils.dump(data));

            fnSuccess();
        },
        function(error)
        {
            fnProgress('Failure during the SRP commit update user');
            fnError(error);
        });
    },

    RollbackUpdateUser: function(newUserVersion, fnProgress, fnSuccess, fnError)
    {
        if (!this.initialized) throw "SRP not initialized";

        fnProgress('SRP rollback update user started');

        fnProgress('Calling the Users/RollbackUpdateUser server\'s method');
        if ($zkBox.logging) $zkBox.utils.log("newUserVersion: " + newUserVersion);
        // call the server's Users/RollbackUpdateUser method
        $zkBox.model.Call(false, true, "DELETE user/update/" + newUserVersion, {},
        function(data)
        {
            fnProgress('The user update was successfully rolledback');
            fnProgress('Decoding the response');

            if ($zkBox.logging) $zkBox.utils.log("data: " + $zkBox.utils.dump(data));

            fnSuccess();
        },
        function(error)
        {
            fnProgress('Failure during the SRP rollback update user');
            fnError(error);
        });
    },    

    DeleteUser: function(login, fnProgress, fnSuccess, fnError)
    {
        if (!this.initialized) throw "SRP not initialized";

        fnProgress('SRP delete user started');

        fnProgress('Calling the Users/DeleteUser server\'s method');
        // call the server's Users/DeleteUser method
        $zkBox.model.Call(false, true, "DELETE user", {},
        function(data)
        {
            fnProgress('The user was successfully deleted');
            fnProgress('Decoding the response');

            if ($zkBox.logging) $zkBox.utils.log("data: " + $zkBox.utils.dump(data));

            fnSuccess();
        },
        function(error)
        {
            fnProgress('Failure during the SRP delete user');
            fnError(error);
        });
    },

    LogoutUser: function(fnProgress, fnSuccess, fnError)
    {
        if (!this.initialized) throw "SRP not initialized";

        fnProgress('SRP logout user started');

        fnProgress('Calling the Users/LogoutUser server\'s method');
        // call the server's Users/LogoutUser method
        $zkBox.model.Call(false, true, "PUT user/logout", {},
        function(data)
        {
            fnProgress('The user was successfully logged out');
            fnProgress('Decoding the response');

            if ($zkBox.logging) $zkBox.utils.log("data: " + $zkBox.utils.dump(data));

            fnSuccess();
        },
        function(error)
        {
            fnProgress('Failure during the SRP logout user');
            fnError(error);
        });
    },

    AddUser: function(username, password, fnProgress, fnSuccess, fnError)
    {
        // Creating a new user
        // username = the username entered in the user interface
        // password = the password entered in the user interface
        if (!this.initialized) throw "SRP not initialized";

        fnProgress('SRP user creation started');
        // login = H(username)
        var login = this.HashUser(username);
        username = '';

        // pass = H(password)
        var pass = this.HashPass(password);
        password = '';
        if ($zkBox.logging) $zkBox.utils.log('passw: ' + pass);

        fnProgress('Salting and hashing the input');
        // s - login's salt
        // s = random number
        var s = randBigInt(this.Nbits, 0); //TODO_LATER: use TRNG
        var sHex = this.to128(this.bigint2hex(s));

        fnProgress('Compute the private key');
        // x - a private key derived from the password and salt
        // x = H(s || H(username) || H(password))
        var x;
        {
            var x1 = this.hex2bigint(this.HHex(
                sHex +
                login +
                pass));
            if (this.compareTo(x1, this.N) < 0) { x = x1; }
            else { x = mod(x1, sub(this.N, this.one)); }
        }

        fnProgress('Generate the server\'s verifier');
        // v - the server's verifier
        // v = g^x (mod N)
        var v = powMod(this.g, x, this.N);
        var vHex = this.to128(this.bigint2hex(v));

        fnProgress('Calling the Users/AddUser server\'s method');
        if ($zkBox.logging) $zkBox.utils.log("login: " + login);
        if ($zkBox.logging) $zkBox.utils.log("    s: " + sHex);
        if ($zkBox.logging) $zkBox.utils.log("    v: " + vHex);
        // call the server's Users/AddUser method
        $zkBox.model.Call(false, false, "POST user", {
            login: login,
            s: sHex,
            v: vHex
        },
        function(data)
        {
            fnProgress('The user was successfully created');
            fnProgress('Decoding the response');

            if ($zkBox.logging) $zkBox.utils.log("data: " + $zkBox.utils.dump(data));

            fnSuccess(
                login,
                pass);
        },
        function(error)
        {
            fnProgress('Failure during the SRP user creation');
            fnError(error);
        });
    },

    AuthenticateUser: function(dataIsHashed, username, password, fnProgress, fnSuccess, fnError)
    {
        // Authenticate an user
        // username = the username entered in the user interface
        // password = the password entered in the user interface
        if (!this.initialized) throw "SRP not initialized";

        fnProgress('SRP authentication started');

        var err = 0;
        var login;
        var pass;

        if (dataIsHashed)
        {
            login = username;
            pass = password;
        }
        else
        {
            // user = H(username)
            login = this.HashUser(username);
            username = '';

            // pass = H(password)
            pass = this.HashPass(password);
            password = '';
        }

        if ($zkBox.logging) $zkBox.utils.log('login: ' + login);
        if ($zkBox.logging) $zkBox.utils.log('passw: ' + pass);

        var hex2bigint = this.hex2bigint;

        // AuthStep1
        var a;
        var A;
        var AHex;
        var uniq;
        var sHex;
        var BHex;
        var u;

        {
            fnProgress('Generate the ephemeral private key');
            // a - ephemeral private key
            // a = random between 2 and N-1
            {
                a = randBigInt(this.Nbits, 0);
                if (this.compareTo(a, this.N) >= 0) { a = mod(a, sub(this.N, this.one)); }
                if (this.compareTo(a, this.two) < 0) { a = this.two; }
            }

            fnProgress('Generate the public key');
            // A - public key
            // A = g^a (mod N)
            A = powMod(this.g, a, this.N);
            AHex = this.to128(this.bigint2hex(A));

            //fnProgress('Calling the Users/AuthStep1 server\'s method');
            fnProgress('Initiating the secure channel');
            if ($zkBox.logging) $zkBox.utils.log("login: " + login);
            if ($zkBox.logging) $zkBox.utils.log("    A: " + AHex);
            // call the server's Users/AuthStep1 method
            $zkBox.model.Call(false, false, "PUT user/auth/step1", {
                login: login,
                A: AHex
            },
            function(data)
            {
                fnProgress('Decoding the response');
                uniq = data.uniq;
                sHex = data.s;
                BHex = data.B;
                u = hex2bigint(data.u);
                if ($zkBox.logging) $zkBox.utils.log(" uniq: " + uniq);
                if ($zkBox.logging) $zkBox.utils.log("    s: " + sHex);
                if ($zkBox.logging) $zkBox.utils.log("    B: " + BHex);
            },
            function(error)
            {
                fnProgress('Failure during the SRP authentication');
                fnError(error);
                err = error;
            });
        }

        // AuthStep2
        var KHex;
        var m1Hex;
        var userId;
        var userVersion;
        var sessionId;
        var m2server;
        if (0 == err)
        {
            var B = this.hex2bigint(BHex)

            // x - private key derived from password and the salt
            // x = H(s || H(username) || H(P))
            var x;
            {
                var x1 = this.hex2bigint(this.HHex(
                    sHex +
                    login +
                    pass));
                if (this.compareTo(x1, this.N) < 0) { x = x1; }
                else { x = mod(x1, sub(this.N, this.one)); }
            }

            fnProgress('Computing SRP common exponential value');
            // S - common exponential value
            // S = (B - k * g^x) ^ (a + u * x) (mod N)
            var S = powMod(mod(sub(add(B, mult(this.N, this.k)), mult(powMod(this.g, x, this.N), this.k)), this.N), add(mult(x, u), a), this.N);

            fnProgress('Computing the strong cryptographically session key');
            // K - the strong cryptographically session key
            // K = H(S)
            var K = this.hex2bigint(this.HHex(this.bigint2hex(S)));
            KHex = this.to128(this.bigint2hex(K));
            if ($zkBox.logging) $zkBox.utils.log("    K: " + KHex);

            fnProgress('Client signing the login session');
            // m1 - client's proof that it has the correct key
            // m1 = H(A, B, K)
            var m1 = this.hex2bigint(this.HHex(
                AHex +
                BHex +
                KHex));
            m1Hex = this.to128(this.bigint2hex(m1));

            //fnProgress('Calling the Users/AuthStep2 server\'s method');
            fnProgress('Establishing the secure connection');
            if ($zkBox.logging) $zkBox.utils.log("login: " + login);
            if ($zkBox.logging) $zkBox.utils.log(" uniq: " + uniq);
            if ($zkBox.logging) $zkBox.utils.log("   m1: " + m1Hex);
            // call the server's Users/AuthStep2 method
            $zkBox.model.Call(false, false, "PUT user/auth/step2", {
                login: login,
                uniq: uniq,
                m1: m1Hex
            },
            function(data)
            {
                fnProgress('Decoding the response');
                userId = data.userId;
                userVersion = data.userVersion;
                sessionId = data.sessionId;

                // m2server - server's proof that it has the correct key
                m2server = hex2bigint(data.m2);

                if ($zkBox.logging) $zkBox.utils.log("useri: " + userId);
                if ($zkBox.logging) $zkBox.utils.log("userv: " + userVersion);
                if ($zkBox.logging) $zkBox.utils.log("sessi: " + sessionId);
                if ($zkBox.logging) $zkBox.utils.log("m2ser: " + data.m2);
            },
            function(error)
            {
                fnProgress('Failure during the SRP authentication');
                fnError(error);
                err = error;
            });
        }

        if (0 == err)
        {
            fnProgress('Checking server\'s signature');
            // m2 - expected server's proof as computed by the client
            // m2 = H(A, m1, K)
            var m2 = this.hex2bigint(this.HHex(
                AHex +
                m1Hex +
                KHex));
            if ($zkBox.logging) $zkBox.utils.log("   m2: " + this.to128(this.bigint2hex(m2)));

            if (0 == this.compareTo(m2, m2server))
            {
                fnProgress('SRP authentication succeded. Computing the crypto key.');
                // the server is trusted
                //if ($zkBox.logging) $zkBox.utils.log('client: the server is trusted');

                var ckey = $zkBox.crypto.cryptokey(pass);
                //if ($zkBox.logging) $zkBox.utils.log(" ckey: " + $zkBox.utils.dump(ckey)); // the key as string
                var cryptokey = base16_encode(str2char(ckey)); // this is displayable
                fnProgress("The crypto key was computed. The secure session was established.");

                if ($zkBox.logging) $zkBox.utils.log("sessionKey: " + KHex);
                if ($zkBox.logging) $zkBox.utils.log(" cryptokey: " + cryptokey);
                //if ($zkBox.logging) $zkBox.utils.log("Dcryptokey: " + base16_decode(cryptokey));
                //if ($zkBox.logging) $zkBox.utils.log("Ucryptokey: " + GibberishAES.enc_utf8(ckey));
                //if ($zkBox.logging) $zkBox.utils.log("U2cryptoke: " + str2utf8(ckey));
                //if ($zkBox.logging) $zkBox.utils.log("Xcryptokey: " + $zkBox.utils.dump(GibberishAES.s2a(ckey)));

                fnSuccess(
                    login,
                    pass,
                    userId,
                    userVersion,
                    sessionId,
                    KHex,
                    cryptokey);
            }
            else
            {
                fnProgress('Failure during the SRP authentication');
                if ($zkBox.logging) $zkBox.utils.log('client: the server is NOT trusted');
                fnError(101); // Invalid username or password
            }
        }
    },

    initialized: false,

    initialize: function(NHex, gHex, kHex)
    {
        this.one = this.hex2bigint("1");
        this.two = this.hex2bigint("2");
        this.three = this.hex2bigint("3");

        this.N = this.hex2bigint(NHex);
        this.Nbits = bitSize(this.N);
        this.g = this.hex2bigint(gHex);
        this.k = this.hex2bigint(kHex);

        this.initialized = true;
    }
};
