﻿// zkBox ASP.NET web proxy
// http://www.zkbox.com/developers/client/javascript/aspnet
// 
// Copyright 2010, Sorin Ostafiev
// http://www.ostafiev.com/
// 
// Released under GPL version 3 license
// http://www.gnu.org/licenses/gpl-3.0.txt


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web;
using System.Net;
using System.IO;
using System.Security.Cryptography;
using Bitground.Common.Helpers;

namespace Bitground.zkBox.Client.JS.WebProxy
{
    public class Handler : IHttpHandler
    {
        private static bool m_initialized = false;
        private static string m_apiUrl = "";
        private static string m_apiVersion = "";
        private static string m_applicationId = "";
        private static string m_applicationKey = "";

        private static Application m_application = null;
        public static Application Application
        {
            get
            {
                if (!m_initialized)
                {
                    throw new CException("Unexpected error: Application was accessed, but the handler was not initialized");
                }

                return m_application;
            }
        }

        public class Init
        {
            public string apiUrl;
            public string apiVersion;
            public string applicationId;
            public string applicationKey;

            internal void CheckParameters()
            {
                Debug.Assert(!String.IsNullOrEmpty(apiUrl), "The initialization parameter apiUrl is null or it wasn't provided");
                Debug.Assert(!String.IsNullOrEmpty(apiVersion), "The initialization parameter apiVersion is null or it wasn't provided");
                Debug.Assert(!String.IsNullOrEmpty(applicationId), "The initialization parameter applicationId is null or it wasn't provided");
                Debug.Assert(!String.IsNullOrEmpty(applicationKey), "The initialization parameter applicationKey is null or it wasn't provided");
            }

            public override string ToString()
            {
                return String.Format(
                    "[apiUrl='{0}' apiVersion='{1}' applicationId='{2}' applicationKey='{3}']",
                    apiUrl,
                    apiVersion,
                    applicationId,
                    applicationKey);
            }
        }

        internal static string LogWebProxyContext(string operation)
        {
            return string.Format(
                "ZKBOX WEBPROXY: {0}",
                operation);
        }

        public static void Initialize(
            Init init)
        {
            string referenceId;
            using (Log.PushRootContext("Global", out referenceId))
            {
                TrueRandom.Initialize();
            }

            if (!m_initialized)
            {
                lock (typeof(Handler))
                {
                    if (!m_initialized)
                    {
                        using (Log.PushRootContext(LogWebProxyContext("START"), out referenceId))
                        {
                            init.CheckParameters();

                            if (Log.CanLogInfo) Log.Info(
                                "Initializing with the following parameters: {0}",
                                init.ToString());

                            m_application = new Application(init);

                            m_apiUrl = init.apiUrl;

                            switch (init.apiVersion)
                            {
                                case "1.0":                 // api version 1.0
                                    m_apiVersion = "0100";
                                    break;

                                default:
                                    throw new ArgumentException("Invalid apiVersion");
                            }

                            m_applicationId = init.applicationId;
                            m_applicationKey = init.applicationKey;

                            // leave this enable to skip the self-signed certificates issues
                            // TODO_LATER: remove this in production
                            // find maybe a better way here to deal with self-signed certificates?
#if DEBUG
                            ServicePointManager.ServerCertificateValidationCallback = delegate { return true; };
#endif

                            m_initialized = true;
                        }
                    }
                }
            }
        }

        public static void Dispose()
        {
            string referenceId;
            using (Log.PushRootContext(LogWebProxyContext("STOP"), out referenceId))
            {
                m_application.Dispose();

                TrueRandom.Dispose();
                if (Log.CanLogInfo) Log.Info("The web proxy was stopped");
            }
        }

        private const string headerZKReference = "X-zkBox-Reference";
        private void SendHttpWebResponse(HttpResponse Response, HttpWebResponse resp)
        {
            using (Log.PushContext("RESPONSE"))
            {
                if (Log.CanLogInfo) Log.Info("Sending back the response to the caller...");

                Response.Clear();
                Response.ClearContent();
                Response.ClearHeaders();

                Response.ContentType = resp.ContentType;
                Response.Charset = resp.CharacterSet;
                Response.StatusCode = (int)resp.StatusCode;
                Response.StatusDescription = resp.StatusDescription;

                foreach (string h in resp.Headers)
                {
                    string hl = h.ToLower();

                    if ((hl != "Content-Type".ToLower()) &&
                        (hl != "Content-Length".ToLower()) &&
                        (hl != "Connection".ToLower()) &&
                        (hl != "Server".ToLower()))
                    {
                        try
                        {
                            if (null == Response.Headers[h])
                            {
                                string val = resp.Headers[h];
                                Response.Headers.Add(h, val);

                                if ((h == headerZKReference) &&
                                    (!String.IsNullOrEmpty(val)))
                                {
                                    if (Log.CanLogInfo) Log.Info("The reference for this call is '{0}'",
                                        val);
                                }

                                if (Log.CanLogTrace) Log.Trace("The header '{0}' (value '{1}') was added to the response", h, val);
                            }
                        }
                        catch(Exception ex)
                        {
                            Log.Warning("Failure while trying to add the header '{0}'. The exception message was '{1}'", h, ex.Message);
                        }
                    }
                }

                using (Stream ist = resp.GetResponseStream())
                {
                    using (StreamReader sr = new StreamReader(ist))
                    {
                        var buff = new char[1024];
                        var length = 0;
                        int chars;
                        while ((chars = sr.Read(buff, 0, 1024)) > 0)
                        {
                            length += chars;
                            Response.Output.Write(buff, 0, chars);
                        }
                        Response.Output.Flush();
                        if (Log.CanLogInfo) Log.Info("The response was sent back");
                    }
                }
            }
        }

        private void SetError(HttpResponse Response, HttpStatusCode statusCode, int subStatusCode)
        {
            Response.ClearContent();
            Response.ClearHeaders();
            Response.Clear();

            Response.SuppressContent = true;

            Response.StatusCode = (int)statusCode;

            if (-1 != subStatusCode)
            {
                Response.SubStatusCode = subStatusCode;
                Log.Warning("Setting the response status code to '{0}' and the sub status code to '{1}'",
                    Response.StatusCode,
                    Response.SubStatusCode);
            }
            else
            {
                Log.Warning("Setting the response status code to '{0}'",
                    Response.StatusCode);
            }

            Response.End();
        }

        private void SetError(HttpResponse Response, HttpStatusCode statusCode)
        {
            SetError(Response, statusCode, -1);
        }

        private void SetInternalError(HttpResponse Response)
        {
            SetError(Response, HttpStatusCode.InternalServerError); // HTTP 500
        }

        private string ByteArrayToHexString(byte[] ar)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < ar.Length; i++)
            {
                sb.Append(ar[i].ToString("X2"));
            }
            return sb.ToString();
        }

        public void ProcessRequest(HttpContext context)
        {
            string referenceId;
            using (Log.PushRootContext(LogWebProxyContext(
                String.Format(
                    "CALL")), out referenceId))
            {

                // sanity checks
                if (null == context)
                {
                    Log.Error("Unexpected error: context is null");
                    return;
                }
                // more sanity checks
                if ((null == context.Request) || (null == context.Response))
                {
                    Log.Error("Unexpected error: context.Request or context.Response is null");
                    return;
                }

                HttpRequest Request = context.Request;
                HttpResponse Response = context.Response;

                // handler not initialized? then abort
                if (!m_initialized)
                {
                    Log.Error("Unexpected error: handler not initialized");
                    SetInternalError(Response);
                    return;
                }

                // not secure connection? then abort
                if (!Request.IsSecureConnection)
                {
                    if (Log.CanLogInfo) Log.Info("Unsecured connection. Abort the call");
                    SetError(Response, HttpStatusCode.Forbidden, 4); // 403.4
                    return;
                }

                int i = Request.Url.AbsoluteUri.IndexOf(Request.CurrentExecutionFilePath);
                if (i < 0)
                {
                    Log.Error("Unexpected error: CurrentExecutionFilePath is missing from AbsoluteUri");
                    SetInternalError(Response);
                    return;
                }

                string resource = Request.Url.AbsoluteUri.Substring(i + Request.CurrentExecutionFilePath.Length);
                string orig = String.Format(
                    "{0}{1}",
                    m_apiUrl,
                    resource);

                if (!(resource.ToLower().StartsWith("/application") ||
                      resource.ToLower().StartsWith("/user") ||
                      resource.ToLower().StartsWith("/object") ||
                      resource.ToLower().StartsWith("/transaction") ||
                      resource.ToLower().StartsWith("/batch")))
                {
                    // we're allowing only the requests for users or objects to pass through
                    if (Log.CanLogInfo) Log.Info("Invalid resource request ({0}). The call will not be forwarded", resource);
                    SetInternalError(Response);
                    return;
                }

                using (Log.PushContext(
                    String.Format(
                        "TO '{0}'",
                        resource)))
                {
                    try
                    {
                        if (Log.CanLogInfo) Log.Info("Preparing the api request...");
                        HttpWebRequest req = HttpWebRequest.Create(
                            new Uri(orig)) as HttpWebRequest;

                        if (Log.CanLogInfo) Log.Info("Fill in the request...");
                        {
                            req.AllowAutoRedirect = false;
                            req.KeepAlive = false;
                            req.Method = Request.HttpMethod;
                            req.UseDefaultCredentials = true;
                            req.UserAgent = Request.UserAgent;
                            if (Request.AcceptTypes.Length > 0)
                            {
                                req.Accept = Request.AcceptTypes[0];
                            }

                            foreach (string h in Request.Headers)
                            {
                                string hl = h.ToLower();

                                if ((hl != "Content-Type".ToLower()) &&
                                    (hl != "Content-Length".ToLower()) &&
                                    (hl != "Connection".ToLower()) &&
                                    (hl != "Accept".ToLower()) &&
                                    (hl != "Referer".ToLower()) &&
                                    (hl != "User-Agent".ToLower()) &&
                                    (hl != "Host".ToLower()))
                                {
                                    if (null == req.Headers[h])
                                    {
                                        try
                                        {
                                            req.Headers.Add(h, Request.Headers[h]);
                                        }
                                        catch
                                        {
                                            Log.Warning("Failure while trying to add the header '{0}'", h);
                                        }
                                    }
                                }
                            }

                            req.ContentType = Request.ContentType;

                            if ((0 < Request.ContentLength) && (null != Request.InputStream))
                            {
                                Stream ist = Request.InputStream;
                                long l = ist.Length;
                                byte[] ar = new byte[l];
                                ist.Read(ar, 0, (int)l);

                                Stream ost = req.GetRequestStream();
                                ost.Write(ar, 0, (int)l);
                                ost.Close();
                            };

                        }

                        if (Log.CanLogInfo) Log.Info("Sign the call...");
                        {
                            string method = Request.HttpMethod;
                            string contentMD5 = (Request.Headers["Content-MD5"] ?? "");
                            string contentType = (Request.Headers["Content-Type"] ?? "");
                            string now = DateTime.UtcNow.ToString("yyyyMMddHHmm");

                            string stringToSign =
                                method + "\n" +
                                resource + "\n" +
                                now + "\n" +
                                contentType + "\n" +
                                contentMD5 + "\n";

                            string signature =
                                ByteArrayToHexString(
                                    new HMACSHA512(
                                        new UTF8Encoding().GetBytes(m_applicationKey)
                                    ).ComputeHash(
                                        new UTF8Encoding().GetBytes(stringToSign)
                                      )
                                );

                            string zkBoxAuth = m_apiVersion + m_applicationId + now + signature;
                            if (Log.CanLogInfo) Log.Info("Setting the X-zkBox-Auth header to '{0}'", zkBoxAuth);
                            req.Headers["X-zkBox-Auth"] = zkBoxAuth;
                        }

                        if (Log.CanLogInfo) Log.Info("Getting the api respounse...");
                        using (HttpWebResponse resp = req.GetResponse() as HttpWebResponse)
                        {
                            SendHttpWebResponse(Response, resp);
                        }
                    }
                    catch (WebException wex)
                    {
                        Log.Exception(wex, "Failure during the call");

                        if ((null != wex.Response) && (wex.Response is HttpWebResponse))
                        {
                            using (HttpWebResponse hwr = wex.Response as HttpWebResponse)
                            {
                                SendHttpWebResponse(Response, hwr);
                            }
                        }
                        else
                        {
                            Log.Error("A web proxy error has occurred");

                            SetInternalError(Response);
                            return;
                        }
                    }
                }
            }
        }

        public bool IsReusable
        {
            get
            {
                return false;
            }
        }
    }
}
